import math
import sys
import hashlib
import os
import glob
import json
import time
from functools import lru_cache
from numbers import Number
from functools import wraps
from itertools import (
    islice,
    chain
)
from typing import (
    Any,
    List,
    Tuple
)

from geopandas.geodataframe import GeoDataFrame
from networkx.classes.multidigraph import MultiDiGraph
import osmnx as ox
import fast_json

from via import logger
from via.settings import (
    MIN_JOURNEY_VERSION,
    MAX_JOURNEY_VERSION
)
from via.constants import (
    METRES_PER_DEGREE,
    DATA_DIR,
    REMOTE_DATA_DIR,
    DEFAULT_EDGE_COLOUR
)
from via.models.gps import GPSPoint
from via.cache_pull import (
    download_cache,
    extract_cache,
    is_cache_already_pulled
)


@lru_cache(maxsize=10000)
def is_journey_data_file(potential_journey_file: str) -> bool:
    """

    :param potential_journey_file:
    :rtype: bool
    :return: if the file contains journey data
    """
    if os.path.splitext(potential_journey_file)[1] != '.json':
        return False

    try:
        with open(potential_journey_file, 'r') as potential_journey_file_io:
            data = fast_json.loads(potential_journey_file_io.read())
    except (json.decoder.JSONDecodeError, ValueError):
        logger.warning('%s json could not be decoded', potential_journey_file)
        return False
    else:
        if not all([
            'uuid' in data,
            'data' in data,
            'transport_type' in data
        ]):
            logger.warning(
                '%s is not a data file as it does not have appropriate data',
                potential_journey_file
            )
            return False

    return True


def get_journeys(
    transport_type=None,
    source=None,
    place=None,
    version_op=None,
    version=None,
    earliest_time=None,
    latest_time=None,
):
    """
    Get local journeys as Journeys

    :kwarg transport_type: bike|car|scooter|whatever else is on the app
    :kwarg source: The data dir to get from
    :kwarg place: The place to get journeys within
    :kwarg version_op: The operator the journey version must pass to
        be returned
    :kwarg version: The version to compare with version_op and the
        journey version
    :rtype: Journey
    """
    from via.models.journeys import Journeys
    return Journeys(
        data=list(
            iter_journeys(
                transport_type=transport_type,
                source=source,
                place=place,
                version_op=version_op,
                version=version,
                earliest_time=earliest_time,
                latest_time=latest_time
            )
        )
    )


def iter_journeys(
    transport_type=None,
    source=None,
    place=None,
    version_op=None,
    version=None,
    earliest_time=None,
    latest_time=None
):
    """
    Get local journeys as iterable of Journey

    :kwarg transport_type: bike|car|scooter|whatever else is on the app
    :kwarg source: The data dir to get from
    :kwarg place: The place to get journeys within
    :kwarg version_op: The operator the journey version must pass to
        be returned
    :kwarg version: The version to compare with version_op and the
        journey version
    """
    from via.models.journey import Journey
    for journey_file in get_data_files(
        transport_type=transport_type,
        source=source
    ):
        journey = Journey.from_file(journey_file)
        if should_include_journey(
            journey,
            place=place,
            version_op=version_op,
            version=version,
            earliest_time=earliest_time,
            latest_time=latest_time
        ):
            yield journey
        else:
            logger.debug('Not including journey: %s', journey)


def should_include_journey(
    journey,
    place=None,
    version_op=None,
    version=None,
    earliest_time=None,
    latest_time=None
):
    if any([
        journey.version < MIN_JOURNEY_VERSION,
        journey.version > MAX_JOURNEY_VERSION
    ]):
        return False

    if version_op is not None:
        if not version_op(journey.version, version):
            return False

    if place is not None:
        if not journey.is_in_place(place):
            return False

    if earliest_time is not None:
        if journey.timestamp < earliest_time:
            return False

    if latest_time is not None:
        if journey.timestamp > latest_time:
            return False

    # is is larger than 50km2
    # in time will split these into smaller journeys internally
    if journey.area > 5e+7:
        return False

    return journey


def get_data_files(transport_type=None, source=None) -> List[str]:
    """

    :kwarg transport_type: bike|car|scooter|whatever else is on the app
    :kwarg source: The data dir to get from
    :rtype: list
    :return: a list of file paths to journey files
    """
    from via.models.journey import Journey
    files = []

    path = DATA_DIR
    if source == 'remote':
        path = REMOTE_DATA_DIR
    elif source is None:
        path = DATA_DIR

    for filename in glob.iglob(path + '/**/*', recursive=True):
        if is_journey_data_file(filename):
            if transport_type not in {None, 'all'}:
                journey_transport_type = os.path.split(os.path.dirname(filename))[-1]
                if journey_transport_type is not None:
                    if journey_transport_type.lower() == transport_type:
                        files.append(filename)
            else:
                files.append(filename)

    return files


def sleep_until_ready(started: Number, finished: Number, max_seconds=0):
    """
    Given a starting time of the kickoff, if there is still time
    to wait, sleep that time, otherwise warn of being overdue.

    :param started:
    :param finished:
    :kwarg max_seconds:
    """
    time_diff = float(finished - started)

    wait_remainder = max_seconds - time_diff
    if wait_remainder > 0:
        time.sleep(wait_remainder)


def timing(function):
    """
    Decorator wrapper to log execution time, for profiling purposes.
    """
    @wraps(function)
    def wrapped(*args, **kwargs):
        start_time = time.monotonic()
        ret = function(*args, **kwargs)
        end_time = time.monotonic()
        logger.debug(
            'timing of "%s"  \t%s',
            function.__qualname__,
            end_time - start_time
        )
        return ret
    return wrapped


def sleep_until(max_seconds: Number):
    """
    Decorator which sleeps until a given x seconds from starting if possible.
    """
    def real_decorator(function):
        def wrapper(*args, **kwargs):
            started = time.monotonic()
            result = function(*args, **kwargs)
            finished = time.monotonic()
            sleep_until_ready(started, finished, max_seconds=max_seconds)

            time_diff = finished - started
            percentage_time = (float(time_diff) / max_seconds) * 100
            if percentage_time > 100:
                # If it's getting up there, then log
                logger.debug(
                    'function "%s" took %s %% of the max %s',
                    function.__qualname__,
                    percentage_time,
                    max_seconds
                )
            return result
        return wrapper
    return real_decorator


def window(sequence, window_size=2):
    """
    Returns a sliding window (of width n) over data from the iterable
    """
    seq_iterator = iter(sequence)
    result = tuple(islice(seq_iterator, window_size))
    if len(result) == window_size:
        yield result
    for elem in seq_iterator:
        result = result[1:] + (elem,)
        yield result


def get_idx_default(lst: list, idx: int, default: Any) -> Any:
    """
    Get the ith elem of a list or a default value if out of range
    """
    assert isinstance(lst, list)
    try:
        return lst[idx]
    except (IndexError, TypeError):
        return default


def get_combined_id(obj: Any, other_obj: Any) -> int:
    """

    :param obj: hashable thing
    :param other_obj: hashable thing
    :rtype: int
    :return: A unque id that will be the same regardless of param order
    """
    return hash(obj) + hash(other_obj)


def get_ox_colours(
    graph: MultiDiGraph,
    colour_map_name: str,
    edge_map=None,
    key_name=None
) -> List:  # TODO: better list hint, not sure what this returns
    """
    NB: only used in local osmnx graphing

    :param graph: MultiDiGraph
    :param colour_map_name: osmnx recognised colour gradient
    :kwarg key_name: The key on the edge data to use as colour intensity data
    :kwarg edge_map: edge_id to data containing colour intensity data
    :rtype: list
    :return: a list of colours indexed by edge order
    """
    if edge_map is not None:
        max_num_colours = max(
            [
                edge_map.get(get_combined_id(u, v), {}).get('avg', -1) for (u, v, _, _) in graph.edges(keys=True, data=True)
            ]
        ) + 1
    elif key_name is not None:
        max_num_colours = max(
            [
                d.get(key_name, -1) for (_, _, _, d) in graph.edges(keys=True, data=True)
            ]
        ) + 1
    else:
        raise Exception('Can not determine what colours to generate. Must give an edge_map or key_name')

    return ox.plot.get_colors(
        n=max_num_colours,
        cmap=colour_map_name
    )


def get_edge_colours(
    graph: MultiDiGraph,
    colour_map_name: str,
    key_name=None,
    edge_map=None
) -> List:  # TODO: better list hint, not sure what this returns
    """
    NB: only used in local osmnx graphing

    :param graph: MultiDiGraph
    :param colour_map_name: osmnx recognised colour gradient
    :kwarg key_name: The key on the edge data to use as colour intensity data
    :kwarg edge_map: edge_id to data containing colour intensity data
    :rtype: list
    :return: a list of colours indexed by edge order
    """
    if key_name is not None:
        colours = get_ox_colours(
            graph,
            colour_map_name,
            key_name=key_name
        )

        return [
            get_idx_default(
                colours,
                d.get(key_name, None),
                DEFAULT_EDGE_COLOUR
            ) for u, v, k, d in graph.edges(
                keys=True,
                data=True
            )
        ]

    if edge_map is not None:
        colours = get_ox_colours(
            graph,
            colour_map_name,
            edge_map=edge_map
        )

        return [
            get_idx_default(
                colours,
                edge_map.get(get_combined_id(u, v), {}).get('avg', None),
                DEFAULT_EDGE_COLOUR
            ) for (u, v, k, d) in graph.edges(
                keys=True,
                data=True
            )
        ]

    raise Exception('Can not determine what colours to generate. Must give an edge_map or key_name')


def force_list(val: Any) -> List[Any]:
    """
    If the val is not already a list, make it the only element in the list

    :rtype: list
    """
    if not isinstance(val, list):
        return [val]
    return val


def flatten(lst: List) -> List[Any]:
    '''
    Given a nested list, flatten it.

    Usage:
        >>> flatten([[1, 2, 3], [1, 2]])
        [1, 2, 3, 1, 2]

    :param lst: list to be flattened
    :return: Flattened list
    '''
    return list(chain.from_iterable(lst))


def filter_nodes_from_geodataframe(
    nodes_dataframe: GeoDataFrame,
    nodes_to_keep: List[int]
) -> GeoDataFrame:
    nodes_to_keep = set(nodes_to_keep)
    to_keep = [
        node for node in nodes_dataframe.index if hash(node) in nodes_to_keep
    ]
    return nodes_dataframe.loc[to_keep]


def filter_edges_from_geodataframe(
    edges_dataframe: GeoDataFrame,
    edges_to_keep: List[Tuple[int, int, int]]
) -> GeoDataFrame:
    edges_to_keep = set([hash(e) for e in edges_to_keep])
    to_keep = [
        edge for edge in edges_dataframe.index if hash(edge) in edges_to_keep
    ]
    return edges_dataframe.loc[to_keep]


def update_edge_data(graph: MultiDiGraph, edge_data_map: dict) -> MultiDiGraph:
    """

    :param graph:
    :param edge_data_map: A dict of values to set on the associated
        edge id (key of edge_data_map)
    """

    # TODO: if fewer items in edge_data_map loop over that
    # rather than graph.edges

    for start, end, _ in graph.edges:
        graph_edge_id = get_combined_id(start, end)
        if graph_edge_id in edge_data_map:
            try:
                graph[start][end][0].update(
                    edge_data_map[graph_edge_id]
                )
            except:
                logger.error(f'Could not update edge: {start} {end}')

    return graph


def get_slope(origin: GPSPoint, dst: GPSPoint):
    return (origin.lng - dst.lng) / (origin.lat - dst.lat)


def get_edge_slope(nodes, edge):
    """
    Given an edge, get the slope of it
    """
    origin = nodes[edge[0][0][0]]
    dst = nodes[edge[0][0][1]]

    origin = GPSPoint(origin['y'], origin['x'])
    dst = GPSPoint(dst['y'], dst['x'])

    return get_slope(origin, dst)


def angle_between_slopes(s1, s2, ensure_positive=False, absolute=False):
    """

    :param s1:
    :param s2:
    :kwargs ensure_positive: Ensure the result is always positive
        Useful in comparisons where you don't care about direction
        and want -45 to also equal 135 for example
    """
    degrees = math.degrees(math.atan((s2 - s1) / (1 + (s2 * s1))))
    if absolute:
        degrees = abs(degrees)
    if ensure_positive:
        if degrees < 0:
            degrees = 180 + degrees
    return degrees


def is_within(bbox, potentially_larger_bbox):
    """

    :param bbox:
    :param potentially_larger_bbox:
    """
    return all([
        bbox['north'] <= potentially_larger_bbox['north'],
        bbox['south'] >= potentially_larger_bbox['south'],
        bbox['east'] <= potentially_larger_bbox['east'],
        bbox['west'] >= potentially_larger_bbox['west']
    ])


def area_from_coords(obj):
    if 'north' in obj.keys():
        vert = obj['north'] - obj['south']
        hori = obj['east'] - obj['west']

        return (vert * METRES_PER_DEGREE) * (hori * METRES_PER_DEGREE)

    else:
        raise NotImplementedError()
        # TODO: Allow for normal polys. Below is an example of doing it for a bbox
        #geom = Polygon(
        #    [
        #        (obj['north'], obj['west']),
        #        (obj['north'], obj['east']),
        #        (obj['south'], obj['east']),
        #        (obj['south'], obj['west']),
        #        (obj['north'], obj['west'])
        #    ]
        #)
        #return ops.transform(
        #    partial(
        #        pyproj.transform,
        #        pyproj.Proj('EPSG:4326'),
        #        pyproj.Proj(
        #            proj='aea',
        #            lat_1=geom.bounds[1],
        #            lat_2=geom.bounds[3]
        #        )
        #    ),
        #    geom
        #).area


@lru_cache(maxsize=5)
def get_graph_id(graph, unreliable=False):
    """

    TODO: have a supporting function that just gets the quick unreliable
          hash so we don't cache the graph itself

    :param graph:
    :kwarg unreliable: use hash which can be different from run to run
    """
    if unreliable:
        return hash(tuple(graph._node))
    return hashlib.md5(str(tuple(graph._node)).encode()).hexdigest()


def get_size(obj, seen=None):
    """Recursively finds size of objects"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size


def setup_cache():
    if not is_cache_already_pulled():
        download_cache()
        extract_cache()



from via.models.generic import (
    GenericObject,
    GenericObjects
)
from via.models.gps import GPSPoint


class Frame(GenericObject):
    """
    A single snapshot of data on a journey containing gps, acceleration
    and time info
    """

    def __init__(self, time, gps, acceleration):
        """

        :param time:
        :param gps: GPSPoint or dict serialization of GPSPoint
        :param acceleration:
        """
        super().__init__()
        self.time = time
        self.gps = GPSPoint.parse(gps)
        self.acceleration = acceleration

    @staticmethod
    def parse(obj):
        if isinstance(obj, dict):
            return Frame(
                obj.get('time', None),
                obj['gps'],
                obj['acc']
            )
        if isinstance(obj, Frame):
            return obj
        raise NotImplementedError(
            'Can\'t parse Frame from type %s' % (type(obj))
        )

    def distance_from(self, point):
        """

        :param point: GPSPoint or tuple of (lat, lng) or Frame object
        :rtype: float
        :return: Distance between points in metres
        """
        if isinstance(point, Frame):
            point = point.gps
        return self.gps.distance_from(point)

    @property
    def is_complete(self):
        """
        Does the frame contain all expected data
        """
        return isinstance(self.time, float) and self.gps.is_populated and self.acceleration != []

    @property
    def road_quality(self):
        return int(self.acceleration.quality * 100)

    def serialize(self, **kwargs):
        data = {
            'gps': self.gps.serialize(),
            'acc': self.acceleration
        }
        if kwargs.get('include_time', True):
            data['time'] = round(self.time, 2)
        return data


class Frames(GenericObjects):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('child_class', Frame)
        super().__init__(*args, **kwargs)

    @property
    def most_northern(self):
        return max([frame.gps.lat for frame in self])

    @property
    def most_southern(self):
        return min([frame.gps.lat for frame in self])

    @property
    def most_eastern(self):
        return max([frame.gps.lng for frame in self])

    @property
    def most_western(self):
        return min([frame.gps.lng for frame in self])

    @property
    def data_quality(self):
        """
        Get the percentage of frames that are good. Should
        automatically disregard journeys with low data quality

        :rtype: float
        :return: The percent between 0 and 1
        """
        # Mixed with the deviation between times?
        return len([f for f in self if f.is_complete]) / float(len(self))

    @property
    def origin(self):
        """

        :rtype: via.models.Frame
        :return: The first frame of the journey
        """
        return self[0]

    @property
    def destination(self):
        """

        :rtype: via.models.Frame
        :return: The last frame of the journey
        """
        return self[-1]

    @property
    def duration(self):
        """

        :rtype: float
        :return: The number of seconds the journey took
        """
        return self.destination.time - self.origin.time

    @property
    def direct_distance(self):
        """

        :rtype: float
        :return: distance from origin to destination in metres
        """
        return self[0].distance_from(self[-1])

    def serialize(self, include_time=True):
        return [frame.serialize(include_time=include_time) for frame in self]



import os
import threading
import datetime
from collections import defaultdict

import fast_json

import numpy as np
from rtree.index import Index as RTreeIndex
from shapely.geometry import Point

from osmnx import utils_graph

from via import logger
from via.settings import VERSION
from via.constants import EDGE_CACHE_DIR
from via.utils import get_combined_id, get_graph_id
from via.bounding_graph_gdfs_cache import utils_bounding_graph_gdfs_cache


GEOM_RTREE_CACHE = defaultdict(dict)


def nearest_edges(G, X, Y, return_dist=False):
    if not (hasattr(X, "__iter__") and hasattr(Y, "__iter__")):
        X = np.array([X])
        Y = np.array([Y])

    if np.isnan(X).any() or np.isnan(Y).any():  # pragma: no cover
        raise ValueError("`X` and `Y` cannot contain nulls")

    graph_id = get_graph_id(G)
    if graph_id in GEOM_RTREE_CACHE:
        geoms = GEOM_RTREE_CACHE[graph_id]['geoms']
        rtree = GEOM_RTREE_CACHE[graph_id]['rtree']
    else:
        if utils_bounding_graph_gdfs_cache.get(get_graph_id(G)) is None:
            utils_bounding_graph_gdfs_cache.set(
                get_graph_id(G),
                utils_graph.graph_to_gdfs(G, nodes=False)["geometry"]
            )

        geoms = utils_bounding_graph_gdfs_cache.get(get_graph_id(G))

        rtree = RTreeIndex()
        for pos, bounds in enumerate(geoms.bounds.values):
            rtree.insert(pos, bounds)

        GEOM_RTREE_CACHE[graph_id] = {
            'geoms': geoms,
            'rtree': rtree
        }

    ne_dist = list()
    for xy in zip(X, Y):
        dists = geoms.iloc[list(rtree.nearest(xy, num_results=12))].distance(Point(xy))

        used_edges = []
        edges = []
        distances = []
        for k, v in dists.to_dict().items():
            if get_combined_id(k[0], k[1]) not in used_edges:
                edges.append(k)
                distances.append(v)
                used_edges.append(get_combined_id(k[0], k[1]))

        ne_dist.append((edges, distances))

    ne, dist = zip(*ne_dist)

    return list(ne), list(dist)



class NearestEdgeCache():
    # TODO: split these in a grid of lat / lng 0.5 by the first gps of the upper right or something. Not very important as files are small
    # TODO: just store uuids as string so no silly casting

    def __init__(self):
        self.loaded = False
        self.data = {}
        self.last_save_len = -1
        self.last_saved_time = datetime.datetime.utcnow()
        self.saver()

    def saver(self):
        self.load()

        if all([
            self.last_save_len < len(self.data),
            (datetime.datetime.utcnow() - self.last_saved_time).total_seconds() > 5
        ]):
            self.save()

        saver = threading.Timer(
            10,
            self.saver
        )
        saver.daemon = True
        saver.start()

    def save(self):
        logger.debug(f'Saving cache {self.fp}')
        with open(self.fp, 'w') as f:
            f.write(fast_json.dumps(self.data))
        self.last_save_len = len(self.data)
        self.last_saved_time = datetime.datetime.utcnow()

    def get(self, graph, frames):
        """

        :param graph:
        :param frames: list of frames to get nearest edges of
        :return: list of list of edge ids or tuple of edge_id and distance.
            Each frame having its own list of closest few roads
        """

        self.load()

        # FIXME: should use frame hash rather than gps hash as we want context too
        frame_ids_to_get = [
            str(frame.gps_hash) for frame in frames if self.data.get(str(frame.gps_hash), None) is None
        ]

        id_frame_map = {str(f.gps_hash): f for f in frames}

        requested_frame_edge_map = {
            str(f.gps_hash): self.data.get(str(f.gps_hash), None) for f in frames
        }

        if frame_ids_to_get != []:
            results = nearest_edges(
                graph,
                [id_frame_map[frame_id].gps.lng for frame_id in frame_ids_to_get],
                [id_frame_map[frame_id].gps.lat for frame_id in frame_ids_to_get],
                return_dist=True
            )

            frame_id_result_map = dict(
                zip(
                    frame_ids_to_get,
                    list(
                        zip(
                            results[0],
                            results[1]
                        )
                    )
                )
            )

            for frame_id, edge_data in frame_id_result_map.items():
                edge_data = list(edge_data)
                self.data[frame_id] = list(zip(edge_data[0], edge_data[1]))

        requested_frame_edge_map = {
            str(f.gps_hash): self.data.get(str(f.gps_hash), None) for f in frames
        }

        return list(requested_frame_edge_map.values())

    def load(self):
        if self.loaded:
            return

        logger.debug(f'Loading cache {self.fp}')
        if not os.path.exists(self.fp):
            os.makedirs(
                os.path.dirname(self.fp),
                exist_ok=True
            )
            self.save()
        with open(self.fp, 'r') as f:
            self.data = fast_json.loads(f.read())
        self.loaded = True
        self.last_save_len = len(self.data)

    @property
    def fp(self):
        # TODO: split by lat lng regions
        return os.path.join(EDGE_CACHE_DIR, VERSION, 'cache.json')


nearest_edge = NearestEdgeCache()



from via.base_cache import BaseCaches, BaseCache
from via.constants import BOUNDING_GRAPH_GDFS_NAME


class UtilsBoundingGraphGDFSGraph(BaseCache):

    def __init__(self, *args, **kwargs):
        kwargs['cache_type'] = 'utils_bounding_graph_gdfs_cache'
        super().__init__(*args, **kwargs)


class UtilsBoundingGraphGDFSGraphs(BaseCaches):

    def __init__(self, *args, **kwargs):
        kwargs['child_class'] = UtilsBoundingGraphGDFSGraph
        kwargs['cache_type'] = 'utils_bounding_graph_gdfs_cache'
        super().__init__(*args, **kwargs)

    def get_fn(self, obj):
        xy = list(obj.to_dict().values())[0].xy
        return '%s_%s.pickle' % (
            round(xy[0][0], 2),
            round(xy[1][0], 2)
        )


class BoundingGraphGDFSGraph(BaseCache):

    def __init__(self, *args, **kwargs):
        kwargs['cache_type'] = BOUNDING_GRAPH_GDFS_NAME
        super().__init__(*args, **kwargs)


class BoundingGraphGDFSGraphs(BaseCaches):

    def __init__(self, *args, **kwargs):
        kwargs['child_class'] = BoundingGraphGDFSGraph
        kwargs['cache_type'] = BOUNDING_GRAPH_GDFS_NAME
        super().__init__(*args, **kwargs)

    def get_fn(self, obj):
        return '%s_%s.pickle' % (
            round(min(obj[0].geometry.x), 2),
            round(min(obj[0].geometry.y), 2)
        )


bounding_graph_gdfs_cache = BoundingGraphGDFSGraphs()
bounding_graph_gdfs_cache.load()

utils_bounding_graph_gdfs_cache = UtilsBoundingGraphGDFSGraphs()
utils_bounding_graph_gdfs_cache.load()



from unittest import TestCase
from mock import patch

from via.base_cache import BaseCache


class BaseCacheTest(TestCase):

    def test_init(self):
        self.assertFalse(
            BaseCache(cache_type='blahblah').loaded
        )
        self.assertEqual(
            BaseCache(cache_type='blahblah').data,
            {}
        )
        self.assertEqual(
            BaseCache(cache_type='blahblah').cache_type,
            'blahblah'
        )

    def test_get(self):
        cache = BaseCache(cache_type='test_get')
        cache.set('one', 1)
        self.assertEqual(
            cache.get('one'),
            1
        )

    def test_save(self):
        cache = BaseCache(cache_type='test_get')
        cache.set('one', 1)
        cache.save()

        cache = BaseCache(cache_type='test_get')
        self.assertEqual(
            cache.get('one'),
            1
        )

    def test_load_content(self):
        cache = BaseCache(cache_type='test_get')
        cache.set('one', 1)
        cache.save()

        cache = BaseCache(cache_type='test_get')
        cache.load()
        self.assertEqual(
            cache.data,
            {
                'one': 1
            }
        )

    def test_load_empty(self):
        cache = BaseCache(cache_type='blahblah')
        cache.load()
        self.assertTrue(cache.loaded)
        self.assertEqual(cache.data, {})

    @patch('via.settings.VERSION', '0.0.1')
    def test_dir(self):
        from via import settings
        self.assertEqual(
            BaseCache(cache_type='blahblah').dir,
            f'/tmp/via/cache/blahblah/{settings.VERSION}'
        )

    @patch('via.settings.VERSION', '0.0.1')
    def test_fp(self):
        from via import settings
        self.assertEqual(
            BaseCache(cache_type='blahblah', fn='cache.pickle').fp,
            f'/tmp/via/cache/blahblah/{settings.VERSION}/cache.pickle'
        )



import hashlib
from numbers import Number
from operator import itemgetter

import numpy
from cached_property import cached_property
from shapely.geometry import (
    MultiPoint,
    Point
)

from via import settings
from via import logger
from via.place_cache import place_cache
from via.models.generic import (
    GenericObject,
    GenericObjects
)
from via.models.gps import GPSPoint
from via.utils import angle_between_slopes


class Context():

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.context_pre = []
        self.context_post = []

    def __del__(self):
        attrs_to_del = [
            'content_hash'
        ]

        for attr in attrs_to_del:
            try:
                delattr(self, attr)
            except:
                pass

    def set_context(self, pre=None, post=None):
        if not self.is_context_populated:
            self.context_pre = pre
            self.context_post = post

    def get_slope_incoming(self, mode='near'):
        """

        :kwarg mode: near or far - how much context should be used
        """
        if mode == 'far':
            return self.context_pre[0].gps.slope_between(self.gps)
        if mode == 'near':
            return self.context_pre[-1].gps.slope_between(self.gps)
        raise ValueError('Mode %s not recognised' % (mode))

    def get_slope_outgoing(self, mode='near'):
        """

        :kwarg mode: near or far - how much context should be used
        """
        if mode == 'far':
            return self.gps.slope_between(self.context_post[-1].gps)
        if mode == 'near':
            return self.gps.slope_between(self.context_post[0].gps)
        raise ValueError(f'Mode {mode} not recognised')

    def get_slope_around_point(self, mode='near'):
        """
        Get the slope of the few points around the point.

        Useful in getting a slope associated with the slope so we can
        compare it to the slope of a road

        :kwarg mode: near or far - how much context should be used
        """
        if mode == 'far':
            return self.context_pre[0].gps.slope_between(
                self.context_post[-1]
            )
        if mode == 'near':
            return self.context_pre[-1].gps.slope_between(
                self.context_post[0]
            )

        raise ValueError('Mode {mode} not recognised')

    def get_in_out_angle(self, mode='near'):
        """

        :kwarg mode: near or far - how much context should be used
        :return: positive smallest angle difference between in slope and
             out slope
        :rtype: float
        """
        # TODO: Rename this function, ew
        return angle_between_slopes(
            self.get_slope_incoming(mode=mode),
            self.get_slope_outgoing(mode=mode),
            absolute=True
        )

    @property
    def is_context_populated(self):
        return self.context_pre != [] and self.context_post != []

    @property
    def dist_to_earliest(self):
        raise NotImplementedError()

    @property
    def dist_to_latest(self):
        raise NotImplementedError()

    def serialize(self, include_time=True):
        return {
            'pre': [p.serialize(include_time=include_time, include_context=False) for p in self.context_pre],
            'post': [p.serialize(include_time=include_time, include_context=False) for p in self.context_post]
        }


class FramePoint(Context, GenericObject):
    """
    Data which snaps to the gps giving something like
    {gps: (1, 2), acc: [1,2,3], time: 1}

    Rather than
    {gps: (1, 2), acc: 1, time: 1}
    {gps: (1, 2), acc: 2, time: 2}
    {gps: (1, 2), acc: 3, time: 3}
    """

    def __init__(self, time, gps, acceleration):
        """

        :param time:
        :param gps: GPSPoint or dict serialization of GPSPoint
        :param acceneration:
        """
        super().__init__()

        self.time = time
        self.gps = GPSPoint.parse(gps)

        if isinstance(acceleration, list):
            self.acceleration = [acc for acc in acceleration if acc >= settings.MIN_ACC_SCORE]
        else:
            if acceleration is None:
                acceleration = 0
            assert isinstance(acceleration, Number)
            self.acceleration = [acc for acc in [acceleration] if acc >= settings.MIN_ACC_SCORE]

    def get_edges_with_context(self, graph, edges):
        """
        Get a list of dicts, giving context to how it relates to the
        current FramePoint

        :param graph:
        :param edges:
        :return: list of dicts with keys
            edge:
            origin: GPSPoint of start of edge
            dst: GPSPoint of end of edge
            slope: the slope of the edge
            angle_between: the smallest angle between the slope of the edge
                and the slope of actual travel
        :rtype: list
        """
        edge_node_data = []
        slope_around = self.get_slope_around_point()

        for edge in edges:
            try:
                origin = graph.nodes[edge[0][0]]
                dst = graph.nodes[edge[0][1]]
                data = {
                    'edge': edge,
                    'origin': GPSPoint(origin['y'], origin['x']),
                    'dst': GPSPoint(dst['y'], dst['x']),
                }
                data['slope'] = data['origin'].slope_between(
                    data['dst']
                )
                data['angle_between'] = angle_between_slopes(
                    slope_around,
                    data['slope'],
                    absolute=True
                )
                edge_node_data.append(
                    data
                )
            except Exception as ex:
                logger.warning('Could not get edge data: %s: %s', edge, ex)

        return edge_node_data

    def get_best_edge(self, edges, graph=None, mode=None):
        """

        :kwarg mode: strategy to use to get the best edge
        """
        default_mode = 'nearest'
        modes_require_graph = {'matching_angle', 'angle_nearest'}
        modes_require_context = {'matching_angle', 'angle_nearest'}

        def nearest():
            return sorted(edges, key=itemgetter(1))[0]

        def matching_angle():
            return sorted(
                self.get_edges_with_context(graph, edges),
                key=lambda x: x['angle_between']
            )[0]['edge']

        def angle_nearest():
            # Find a middleground between the best angle match and the
            # nearest by distance
            edges_by_angle = sorted(
                self.get_edges_with_context(graph, edges),
                key=lambda x: x['angle_between']
            )

            best_edge = edges_by_angle[0]
            for edge in edges_by_angle:
                # if best angle match is x degrees within the next and the
                # next is closer, use the closer one
                if all([
                    edge['angle_between'] < settings.CLOSE_ANGLE_TO_ROAD,
                    edge['edge'][1] < best_edge['edge'][1]
                ]):
                    best_edge = edge

            return best_edge['edge']

        if mode is None:
            mode = default_mode

        if mode in modes_require_graph:
            if not graph:
                logger.warning('graph not supplied to get_best_edge and mode \'%s\' was selected. Defaulting to mode \'%s\'', mode, default_mode)
                return self.get_best_edge(edges, mode=default_mode)

        if mode in modes_require_context:
            if not self.is_context_populated:
                logger.debug('Cannot use mode \'%s\' as point context is not populated, using mode \'%s\'', mode, default_mode)
                # can probably warn if there's no post AND no pre, that would
                # show there was no context ever set on the journey?
                return self.get_best_edge(edges, mode='nearest')

        # Remove footway (unless there's no other options).
        # May want to keep included if it's the only thing close
        if graph is not None:

            highways_to_exclude = {
                'footway',
                'steps',
                'corridor',
                'sidewalk',
                'crossing'
            }

            without_footway = []
            for edge in edges:
                highway = graph.edges[edge[0]]['highway']
                include = True

                if not isinstance(highway, list):
                    include = highway not in highways_to_exclude

                # TODO: if highway is a list be more fancy

                if include:
                    without_footway.append(edge)

            if without_footway != []:
                edges = without_footway

        if mode == 'nearest':
            return nearest()
        if mode == 'matching_angle':
            return matching_angle()
        if mode == 'angle_nearest':
            return angle_nearest()
        if mode == 'sticky':
            # Try to stick to previous road if it makes sense
            # Might want to be sticky on top of some other mode?
            # Not important now
            raise NotImplementedError()

        logger.warning('Can not use mode \'%s\' to get best edge as that is not recognised. Defaulting to mode \'%s\'', mode, default_mode)
        return self.get_best_edge(edges, mode=default_mode)

    def append_acceleration(self, acc):
        if isinstance(acc, list):
            for item in acc:
                self.append_acceleration(item)
        elif isinstance(acc, type(None)):
            return
        else:
            if acc >= settings.MIN_ACC_SCORE:
                self.acceleration.append(acc)

    @staticmethod
    def parse(obj):
        if isinstance(obj, FramePoint):
            return obj

        if isinstance(obj, dict):
            return FramePoint(
                obj.get('time', None),
                obj['gps'],
                [acc for acc in obj['acc'] if acc >= settings.MIN_ACC_SCORE]
            )
        raise NotImplementedError(
            f'Can\'t parse Point from type {type(obj)}'
        )

    def speed_between(self, point):
        """
        Get the speed between this and another point (as the crow flies) in
        metres per second

        :param point: A FramePoint obj
        :return: metres per second
        :rtype: float
        """
        metres_per_second = None
        distance = self.distance_from(point.gps)
        if distance != 0:
            time_diff = point.time - self.time
            metres_per_second = distance / time_diff
        return metres_per_second

    def distance_from(self, point):
        """

        :param point: GPSPoint or tuple of (lat, lng) or Frame object
        :rtype: float
        :return: Distance between points in metres
        """
        if isinstance(point, FramePoint):
            point = point.gps
        return self.gps.distance_from(point)

    @property
    def is_complete(self):
        """
        Does the point contain all expected data
        """
        return isinstance(self.time, float) and self.gps.is_populated and self.acceleration != []

    @property
    def road_quality(self):
        """
        Get the average quality at this point (and a little before)

        :return: mean of acceleration points
        :rtype: float
        """
        if self.acceleration == []:
            return 0
        try:
            return int(numpy.mean(self.acceleration) * 100)
        except:
            logger.warning(f'Could not calculate road quality from: {self.acceleration}. Defauling to 0')
            return 0

    def serialize(self, include_time=True, include_context=True):
        data = {
            'gps': self.gps.serialize(),
            'acc': list(self.acceleration),
        }
        if include_time:
            data['time'] = round(self.time, 2)
        if include_context:
            data['context'] = {
                'pre': [p.serialize(include_time=include_time, include_context=False) for p in self.context_pre],
                'post': [p.serialize(include_time=include_time, include_context=False) for p in self.context_post],
            }
        return data

    @property
    def gps_hash(self):
        return self.gps.content_hash

    @cached_property
    def content_hash(self):
        return int.from_bytes(f'{self.acceleration} {self.gps.point} {self.time}'.encode(), 'little') % 2**100


class FramePoints(GenericObjects):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('child_class', FramePoint)
        super().__init__(*args, **kwargs)

    def __del__(self):
        attrs_to_del = [
            'country'
        ]

        for attr in attrs_to_del:
            try:
                delattr(self, attr)
            except:
                pass

    @property
    def most_northern(self):
        return max([frame.gps.lat for frame in self])

    @property
    def most_southern(self):
        return min([frame.gps.lat for frame in self])

    @property
    def most_eastern(self):
        return max([frame.gps.lng for frame in self])

    @property
    def most_western(self):
        return min([frame.gps.lng for frame in self])

    @property
    def bbox(self):
        return {
            'north': self.most_northern,
            'south': self.most_southern,
            'east': self.most_eastern,
            'west': self.most_western,
        }

    @property
    def data_quality(self):
        """
        Get the percentage of frames that are good. Should
        automatically disregard journeys with low data quality

        :rtype: float
        :return: The percent between 0 and 1
        """
        return 1  # TODO

    @property
    def origin(self):
        """

        :rtype: via.models.Frame
        :return: The first frame of the journey
        """
        return self[0]

    @property
    def destination(self):
        """

        :rtype: via.models.Frame
        :return: The last frame of the journey
        """
        return self[-1]

    @property
    def duration(self):
        """

        :rtype: float
        :return: The number of seconds the journey took
        """
        return self.destination.time - self.origin.time

    @property
    def direct_distance(self):
        """

        :rtype: float
        :return: distance from origin to destination in metres
        """
        return self[0].distance_from(self[-1])

    def serialize(self, include_time=False, include_context=True):
        return [
            frame.serialize(
                include_time=include_time,
                include_context=include_context
            ) for frame in self
        ]

    def get_multi_points(self):
        """
        Get a shapely.geometry.MultiPoint of all the points
        """
        unique_points = []
        prev = None
        for frame in self:
            if frame.gps.is_populated:
                if prev is not None:
                    if prev.gps.lat != frame.gps.lat:
                        unique_points.append(
                            Point(
                                frame.gps.lng,
                                frame.gps.lat
                            )
                        )

            prev = frame

        return MultiPoint(
            unique_points
        )

    @property
    def gps_hash(self):
        return hashlib.md5(
            str([
                point.gps.content_hash for point in self
            ]).encode()
        ).hexdigest()

    @property
    def content_hash(self):
        return hashlib.md5(
            str([
                point.content_hash for point in self
            ]).encode()
        ).hexdigest()

    @cached_property
    def country(self):
        """
        Get what country this journey started in

        :return: a two letter country code
        :rtype: str
        """
        return self.origin.gps.reverse_geo['cc']

    def is_in_place(self, place_name: str):
        """
        Get if a journey is entirely within the bounds of some place.
        Does this by rect rather than polygon so it isn't exact but mainly
        to be used to focus on a single city.

        :param place_name: An osmnx place name. For example "Dublin, Ireland"
            To see if the place name is valid try graph_from_place(place).
            Might be good to do that in here and throw an ex if it's not found
        """
        return place_cache.is_in_place(self.bbox, place_name)



import os
from collections import defaultdict

import fast_json
import osmnx as ox

from via import settings
from via import logger
from via.utils import (
    filter_nodes_from_geodataframe,
    filter_edges_from_geodataframe,
    update_edge_data,
    get_graph_id,
    area_from_coords
)
from via.nearest_edge import nearest_edge
from via.network_cache import network_cache
from via.constants import GEOJSON_DIR

from via.bounding_graph_gdfs_cache import bounding_graph_gdfs_cache

from via.geojson.utils import geojson_from_graph


class SnappedRouteGraphMixin():

    @property
    def snapped_route_graph(self):
        """
        Get the route graph, snapping to the nearest edge
        """
        bounding_graph = self.graph

        nearest_edge.get(bounding_graph, self.all_points)

        edges = []
        used_node_ids = []

        all_nearest_edges = nearest_edge.get(
            bounding_graph,
            self.all_points
        )

        for our_origin, nearest_edges in list(zip(self.all_points, all_nearest_edges)):
            edge = our_origin.get_best_edge(
                nearest_edges,
                mode=settings.NEAREST_EDGE_METHOD,
                graph=bounding_graph
            )

            edges.append(tuple(edge[0]))
            used_node_ids.extend([edge[0][0], edge[0][1]])

        if bounding_graph_gdfs_cache.get(get_graph_id(bounding_graph)) is None:
            bounding_graph_gdfs_cache.set(
                get_graph_id(bounding_graph),
                ox.graph_to_gdfs(
                    bounding_graph,
                    fill_edge_geometry=True
                )
            )

        graph_nodes, graph_edges = bounding_graph_gdfs_cache.get(get_graph_id(bounding_graph))

        # Filter only the nodes and edges on the route and ignore the
        # buffer used to get context
        graph = ox.graph_from_gdfs(
            filter_nodes_from_geodataframe(graph_nodes, used_node_ids),
            filter_edges_from_geodataframe(graph_edges, edges)
        )

        return update_edge_data(graph, self.edge_quality_map)


class GeoJsonMixin():

    def save_geojson(self):
        geojson_file = os.path.join(
            GEOJSON_DIR,
            self.content_hash + '.geojson'
        )

        if not os.path.exists(geojson_file):
            logger.debug(f'{geojson_file} not found, generating...')

            if not os.path.exists(geojson_file):
                os.makedirs(
                    os.path.join(GEOJSON_DIR),
                    exist_ok=True
                )

            with open(geojson_file, 'w') as json_file:
                fast_json.dump(
                    self.geojson,
                    json_file,
                    indent=4
                )

    @property
    def geojson(self):
        """
        Write and return a GeoJSON object string of the graph.
        """

        geojson_file = os.path.join(
            GEOJSON_DIR,
            self.content_hash + '.geojson'
        )

        logger.debug(f'{geojson_file} generating...')

        from via.models.journeys import Journeys

        if isinstance(self, Journeys):
            region_map = defaultdict(Journeys)

            for journey in self:
                # use place_2 as place_1 is too specific and a journey that
                # starts in a place_1 could share roads with a journey that
                # starts in a different area nearby
                # This is also a possible issue with place_2 but will happen
                # much less, still a FIXME
                region_name = journey.origin.gps.reverse_geo['place_2']
                region_map[region_name].append(journey)

            if len(region_map) > 1:
                geo_features = []
                for k, v in region_map.items():
                    logger.debug(f'Getting geojson features of journeys group in region: {k}')
                    geo_features.extend(v.geojson['features'])

                geo_features = {
                    'type': 'FeatureCollection',
                    'features': geo_features
                }
                return geo_features
            else:
                return geojson_from_graph(self.snapped_route_graph)

        return geojson_from_graph(self.snapped_route_graph)


class BoundingGraphMixin():

    def get_bounding_graph(self, use_graph_cache=True):
        logger.debug(
            'Plotting bounding graph (n,s,e,w) (%s, %s, %s, %s)',
            self.most_northern,
            self.most_southern,
            self.most_eastern,
            self.most_western
        )

        if use_graph_cache is False or network_cache.get('bbox', self) is None:
            logger.debug(f'bbox > {self.gps_hash} not found in cache, generating...')
            network = ox.graph_from_bbox(
                self.most_northern,
                self.most_southern,
                self.most_eastern,
                self.most_western,
                network_type=self.network_type,
                simplify=True
            )
            network_cache.set(
                'bbox',
                self,
                network
            )

        return network_cache.get('bbox', self)

    @property
    def bounding_graph(self):
        """
        Get a rectangular graph which contains the journey
        """
        return self.get_bounding_graph(use_graph_cache=True)

    @property
    def area(self):
        return area_from_coords(self.bbox)



import os
import time
from shutil import (
    copyfile,
    rmtree
)

from unittest import TestCase, skip

import geopandas

from via.utils import (
    get_slope,
    get_edge_slope,
    get_ox_colours,
    is_journey_data_file,
    get_data_files,
    window,
    iter_journeys,
    sleep_until,
    get_idx_default,
    force_list,
    flatten,
    get_journeys,
    angle_between_slopes,
    get_slope,
    filter_edges_from_geodataframe,
    filter_nodes_from_geodataframe
)
from via.constants import (
    REMOTE_DATA_DIR,
    DATA_DIR
)
from via.models.gps import GPSPoint


@sleep_until(0.5)
def sleep_a_bit():
    return None


class UtilTest(TestCase):

    def setUp(self):
        try:
            rmtree(DATA_DIR)
        except Exception as ex:
            pass
        os.makedirs(REMOTE_DATA_DIR, exist_ok=True)
        copyfile(
            'test/resources/journey_point_at_node.json',
            os.path.join(REMOTE_DATA_DIR, 'journey_point_at_node.json')
        )
        copyfile(
            'test/resources/just_route.json',
            os.path.join(REMOTE_DATA_DIR, 'just_route.json')
        )

    def test_is_journey_data_file(self):
        self.assertTrue(
            is_journey_data_file(
                os.path.join(
                    REMOTE_DATA_DIR,
                    'journey_point_at_node.json'
                )
            )
        )

        self.assertFalse(
            is_journey_data_file(
                os.path.join(
                    REMOTE_DATA_DIR,
                    'just_route.json'
                )
            )
        )

        self.assertFalse(
            is_journey_data_file(
                '/dev/null'
            )
        )

    def test_get_journeys(self):
        self.assertEqual(
            len(get_journeys()),
            1
        )


    def test_window(self):
        self.assertEqual(
            list(window([1, 2, 3, 4, 5])),
            [(1, 2), (2, 3), (3, 4), (4, 5)]
        )

        self.assertEqual(
            list(window([1, 2, 3, 4, 5], window_size=2)),
            [(1, 2), (2, 3), (3, 4), (4, 5)]
        )

        self.assertEqual(
            list(window([1, 2, 3, 4, 5], window_size=3)),
            [(1, 2, 3), (2, 3, 4), (3, 4, 5)]
        )

    def test_sleep_until(self):
        st = time.monotonic()
        sleep_a_bit()
        et = time.monotonic()
        taken = et - st

        self.assertTrue(taken >= 0.5)

    def test_get_idx_default(self):
        self.assertEqual(
            get_idx_default([1, 2, 3], 0, None),
            1
        )
        self.assertEqual(
            get_idx_default([1, 2, 3], 9, None),
            None
        )

    def test_force_list(self):
        self.assertEqual(
            force_list([]),
            []
        )
        self.assertEqual(
            force_list([[]]),
            [[]]
        )
        self.assertEqual(
            force_list(1),
            [1]
        )

    def test_flatten(self):
        self.assertEqual(
            flatten([[1, 2, 3], [1, 2]]),
            [1, 2, 3, 1, 2]
        )

    @skip('todo')
    def test_get_ox_colours(self):
        # TODO: once not doing random quality
        pass

    @skip('todo')
    def test_get_edge_colours(self):
        # TODO: once not doing random quality
        pass

    @skip('todo')
    def test_update_edge_data(self):
        pass

    @skip('todo')
    def test_get_edge_angle(self):
        pass

    def test_angle_between_slopes(self):
        self.assertEqual(
            angle_between_slopes(0, 1),
            45
        )
        self.assertEqual(
            angle_between_slopes(1, 0),
            -45
        )
        self.assertEqual(
            angle_between_slopes(0, 1, ensure_positive=True),
            45
        )
        self.assertEqual(
            angle_between_slopes(1, 0, ensure_positive=True),
            135
        )

    def test_get_slope(self):
        self.assertEqual(
            get_slope(
                GPSPoint(0, 0),
                GPSPoint(1, 1),
            ),
            1
        )

        self.assertEqual(
            get_slope(
                GPSPoint(1, 1),
                GPSPoint(0, 0),
            ),
            1
        )

    def test_get_edge_slope(self):
        self.assertEqual(
            get_edge_slope(
                {1: {'x': 0, 'y': 0}, 2: {'x': 1, 'y': 1}},
                [[[1, 2]]]
            ),
            1
        )

    def test_filter_edges_from_geodataframe(self):
        json_dataframe = {'osmid': {(385708, 446897184, 0): 874491383, (385713, 325240800, 0): 22747810, (385713, 356046111, 0): 25882052, (385713, 325241075, 0): 72248071}, 'oneway': {(385708, 446897184, 0): True, (385713, 325240800, 0): False, (385713, 356046111, 0): False, (385713, 325241075, 0): False}, 'lanes': {(385708, 446897184, 0): '4', (385713, 325240800, 0): '2', (385713, 356046111, 0): None, (385713, 325241075, 0): '3'}, 'ref': {(385708, 446897184, 0): 'M50', (385713, 325240800, 0): None, (385713, 356046111, 0): None, (385713, 325241075, 0): 'L3020'}, 'name': {(385708, 446897184, 0): 'Northern Cross Route Motorway', (385713, 325240800, 0): 'Mill Road', (385713, 356046111, 0): 'Church Avenue', (385713, 325241075, 0): 'Main Street'}}
        df = geopandas.geodataframe.DataFrame.from_dict(json_dataframe)

        self.assertEqual(
            [o['osmid'] for o in filter_edges_from_geodataframe(df, [(385708, 446897184, 0), (385713, 325240800, 0)]).to_dict(orient='records')],
            [874491383, 22747810]
        )

    def filter_nodes_from_geodataframe(self):
        json_dataframe = {'y': {385708: 53.3888025, 385713: 53.3858427, 385738: 53.3890498, 385745: 53.3949888}, 'x': {385708: -6.3537148, 385713: -6.3747161, 385738: -6.3758631, 385745: -6.3751901}, 'street_count': {385708: 3, 385713: 4, 385738: 3, 385745: 3}}
        df = geopandas.geodataframe.DataFrame.from_dict(json_dataframe)

        self.assertEqual(
            list([o['osmid'] for o in filter_nodes_from_geodataframe(df, [385708, 385713]).to_dict(orient='records')].index),
            [385708, 385713]
        )



import statistics
import hashlib
import multiprocessing
from contextlib import closing
from collections import defaultdict
from typing import List

import osmnx as ox
from networkx.classes.multidigraph import MultiDiGraph

from via.models.generic import GenericObjects
from via import logger
from via.models.journey import Journey
from via.utils import (
    get_edge_colours,
    flatten
)
from via.models.journey_mixins import (
    SnappedRouteGraphMixin,
    GeoJsonMixin,
    BoundingGraphMixin
)


def get_journey_edge_quality_map(journey):
    edge_quality_map = defaultdict(list)
    for edge_hash, edge_quality in journey.edge_quality_map.items():
        edge_quality_map[edge_hash].append(edge_quality)

    return dict(edge_quality_map)


class Journeys(
    GenericObjects,
    SnappedRouteGraphMixin,
    GeoJsonMixin,
    BoundingGraphMixin
):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('child_class', Journey)
        super().__init__(*args, **kwargs)

        network_types = [journey.network_type for journey in self]
        if len(set(network_types)) == 0 or len(set(network_types)) > 1:
            self.network_type = 'all'
        elif len(set(network_types)) == 1:
            self.network_type = network_types[0]

    def plot_routes(
        self,
        use_graph_cache=False,
        use_closest_edge_from_base=False,
        min_edge_usage=1,
        colour_map_name='bwr',
        plot_kwargs=None
    ):
        """

        :kwarg use_closest_from_base: For each point on the actual route, for
            each node use the closest node from the original base graph
            the route is being drawn on
        :kwargs min_edge_usage: The minimum number of times an edge has to be
            used for it to be included in the final data (1 per journey)
        :kwarg colour_map_name:
        :kwarg plot_kwargs: A dict of kwargs to pass to whatever plot is
            being done
        """
        if plot_kwargs is None:
            plot_kwargs = {}
        if len(self) == 0:
            raise Exception('Current Journeys object has no content')
        if len(self) == 1:
            logger.warning('To use Journeys effectively multiple journeys must be used, only one found')

        base = self.get_graph(use_graph_cache=use_graph_cache)
        if use_closest_edge_from_base:
            edge_colours = get_edge_colours(
                base,
                colour_map_name,
                edge_map={
                    edge_id: data for edge_id, data in self.edge_quality_map.items() if data['count'] >= min_edge_usage
                }
            )
        else:
            for journey in self:
                base.add_nodes_from(journey.route_graph.nodes(data=True))
                base.add_edges_from(journey.route_graph.edges(data=True))

            edge_colours = get_edge_colours(
                base,
                colour_map_name,
                key_name='avg_road_quality'
            )

        ox.plot_graph(
            base,
            edge_color=edge_colours,
            **plot_kwargs
        )

    def get_mega_journeys(self):
        megas = defaultdict(Journey)

        for journey in self:
            key = '%s_%s' % (journey.transport_type, journey.suspension)
            megas[key].extend([frame for frame in journey])
            megas[key].transport_type = journey.transport_type
            megas[key].suspension = journey.suspension
            megas[key].included_journeys.append(journey)

        return megas

    @property
    def edge_quality_map(self):
        """
        Get a map between edge_hash and road quality of the road. edge_map
        being edge id and road quality being something that hasn't been
        defined yet TODO

        :rtype: dict
        """

        #with multiprocessing.Pool(min([multiprocessing.cpu_count() - 1, 4])) as pool:
        #    journey_edge_quality_maps = pool.map(
        #        get_journey_edge_quality_map,
        #        self
        #    )

        journey_edge_quality_maps = [get_journey_edge_quality_map(i) for i in self]

        edge_quality_map = defaultdict(list)
        for journey_edge_quality_map in journey_edge_quality_maps:
            for edge_id, quals in journey_edge_quality_map.items():
                edge_quality_map[edge_id].extend(quals)

        return {
            edge_id: {
                'avg': int(statistics.mean([d['avg'] for d in data])),
                'count': len(data),
                'accidents': 0,
            } for edge_id, data in edge_quality_map.items()
        }

    @property
    def most_northern(self) -> float:
        """
        Get the most northerly latitude over all journeys
        """
        return max([journey.most_northern for journey in self])

    @property
    def most_southern(self) -> float:
        """
        Get the most southerly latitude over all journeys
        """
        return min([journey.most_southern for journey in self])

    @property
    def most_eastern(self) -> float:
        """
        Get the most easterly longitude over all journeys
        """
        return max([journey.most_eastern for journey in self])

    @property
    def most_western(self) -> float:
        """
        Get the most westerly longitude over all journeys
        """
        return min([journey.most_western for journey in self])

    def get_graph(self, use_graph_cache=True):
        return self.get_bounding_graph(use_graph_cache=use_graph_cache)

    @property
    def graph(self) -> MultiDiGraph:
        return self.get_graph(use_graph_cache=True)

    @staticmethod
    def from_files(filepaths):
        with closing(multiprocessing.Pool(multiprocessing.cpu_count() - 1)) as pool:
            journeys = list(pool.imap_unordered(Journey.from_file, filepaths))
        return Journeys(
            data=journeys
        )

    @property
    def gps_hash(self) -> str:
        return hashlib.md5(
            str([
                journey.gps_hash for journey in self
            ]).encode()
        ).hexdigest()

    @property
    def content_hash(self) -> str:
        return hashlib.md5(
            str([
                journey.content_hash for journey in self
            ]).encode()
        ).hexdigest()

    @property
    def all_points(self) -> List:
        """
        Return all the points in this journeys obj
        """
        return flatten([journey.all_points for journey in self._data])

    @property
    def bbox(self):
        return {
            'north': self.most_northern,
            'south': self.most_southern,
            'east': self.most_eastern,
            'west': self.most_western
        }



import requests

import warnings
warnings.filterwarnings('ignore', '.*Geometry is in a geographic CRS.*', )

import osmnx as ox

from via import settings
from via.log import logger


# overpass api clone for this so we can disable rate limiting and not feel bad

try:
    CUSTOM_AVAILABLE = 'www.openstreetmap.org' in requests.get(settings.CUSTOM_OVERPASS_API, timeout=1).text
except:
    CUSTOM_AVAILABLE = False

if CUSTOM_AVAILABLE:
    overpass_endpoint = settings.CUSTOM_OVERPASS_API
    overpass_rate_limit = False
    logger.info('Using custom overpass api')
else:
    overpass_endpoint = settings.DEFAULT_OVERPASS_API
    overpass_rate_limit = True
    logger.info('Using default overpass api')

ox.config(
    overpass_endpoint=overpass_endpoint,
    overpass_rate_limit=overpass_rate_limit
)



import os
import pkg_resources
from packaging import version

# TODO: data from DOWNLOAD_JOURNEYS_URL should give back the s3 region
DOWNLOAD_JOURNEYS_URL = os.getenv('DOWNLOAD_JOURNEYS', 'https://l7vv5djf9h.execute-api.eu-west-1.amazonaws.com/default/getUUIDs')
S3_REGION = os.getenv('S3_REGION', 'eu-west-1')

PREPARED_GEOJSON_BUCKET = os.getenv('PREPARED_GEOJSON_BUCKET', 'via-geojson')
PREPARED_CACHE_BUCKET = os.getenv('PREPARED_CACHE_BUCKET', 'via-cache')

MIN_ACC_SCORE = float(os.getenv('MIN_ACC_SCORE', '0.001'))
MIN_PER_JOURNEY_USAGE = float(os.getenv('MIN_PER_JOURNEY_USAGE', '1'))
MIN_METRES_PER_SECOND = float(os.getenv('MIN_METRES_PER_SECOND', '2'))
MAX_METRES_PER_SECOND = float(os.getenv('MAX_METRES_PER_SECOND', '10000'))

MIN_JOURNEY_VERSION = version.parse(os.getenv('MIN_JOURNEY_VERSION', '0.0.0'))
MAX_JOURNEY_VERSION = version.parse(os.getenv('MAX_JOURNEY_VERSION', '999.999.999'))

# How often to skip gps points, smooths things out a bit more
# 1 includes all, 3 includes every 3rd etc
# Intervals are generally 2 seconds
GPS_INCLUDE_RATIO = int(os.getenv('GPS_INCLUDE_RATIO', '2'))

NEAREST_EDGE_METHOD = os.getenv('NEAREST_EDGE_METHOD', 'angle_nearest')
CLOSE_ANGLE_TO_ROAD = float(os.getenv('CLOSE_ANGLE_TO_ROAD', 15))

DEFAULT_OVERPASS_API = os.getenv('DEFAULT_OVERPASS_API', 'https://overpass-api.de/api')
CUSTOM_OVERPASS_API = os.getenv('CUSTOM_OVERPASS_API', 'http://54.73.95.15/api')

ENABLE_COLLISIONS = True if os.getenv('ENABLE_COLLISIONS', 'False').lower()[0] == 't' else False

CLEAN_MEMORY = os.getenv('TEST_ENV', 'False') == 'False'

VERSION = pkg_resources.require('via-api')[0].version



import os

METRES_PER_DEGREE = 110574.38855780

BASE_DIR = '/opt/via/' if os.getenv('TEST_ENV', 'False') == 'False' else '/tmp/via'

DATA_DIR = os.path.join(BASE_DIR, 'data')
GENERATED_DIR = os.path.join(BASE_DIR, 'generated')
CACHE_DIR = os.path.join(BASE_DIR, 'cache')

REMOTE_DATA_DIR = os.path.join(DATA_DIR, 'raw')

EDGE_CACHE_NAME = 'edge_cache'
NETWORK_CACHE_NAME = 'network_cache'
BOUNDING_GRAPH_GDFS_NAME = 'bounding_graph_gdfs_cache'

EDGE_CACHE_DIR = os.path.join(CACHE_DIR, EDGE_CACHE_NAME)
NETWORK_CACHE_DIR = os.path.join(CACHE_DIR, NETWORK_CACHE_NAME)
BOUNDING_GRAPH_GDFS_CACHE = os.path.join(CACHE_DIR, BOUNDING_GRAPH_GDFS_NAME)

GEOJSON_DIR = os.path.join(GENERATED_DIR, 'geojson')

LOG_LOCATION = '/var/log/via/via.log' if os.getenv('TEST_ENV', 'False') == 'False' else '/tmp/log/via/via.log'

DEFAULT_EDGE_COLOUR = '#E1E1E1'

POLY_POINT_BUFFER = 0.002



import os
import re
import zipfile
import shutil
from packaging import version

import boto3
from botocore import UNSIGNED
from botocore.client import Config

from via import logger
from via.settings import (
    S3_REGION,
    PREPARED_CACHE_BUCKET,
    VERSION
)
from via.constants import (
    BASE_DIR,
    CACHE_DIR,
    EDGE_CACHE_DIR,
    NETWORK_CACHE_DIR,
    BOUNDING_GRAPH_GDFS_CACHE
)


def download_cache():
    logger.info('Downloading cache')
    os.makedirs(
        CACHE_DIR,
        exist_ok=True
    )

    s3 = boto3.client(
        's3',
        region_name=S3_REGION,
        config=Config(signature_version=UNSIGNED)
    )

    available_versions = []

    paginator = s3.get_paginator('list_objects')
    for result in paginator.paginate(Bucket=PREPARED_CACHE_BUCKET):
        for obj in result.get('Contents', []):
            key = obj['Key']
            try:
                available_versions.append(
                    (
                        version.parse(os.path.splitext(key)[0]),
                        key
                    )
                )
            except:
                pass

    sorted_versions = sorted(
        available_versions,
        key=lambda x: x[1],
        reverse=True
    )
    valid_versions = [
        v for v in sorted_versions if v[0] <= version.parse(VERSION)
    ]

    pull_version = valid_versions[0]

    if pull_version[0] != version.parse(VERSION):
        logger.warning(
            'Pulling the most recent (older) cache which may not be compatible. Current %s Pulling %s',
            VERSION,
            pull_version[0]
        )

    s3.download_file(
        PREPARED_CACHE_BUCKET,
        pull_version[1],
        os.path.join(BASE_DIR, f'cache_{pull_version[1]}')
    )


def extract_cache():
    logger.info('Extracting cache')
    cache_files = [
        f for f in os.listdir(BASE_DIR) if all([
            f.startswith('cache_'),
            f.endswith('.zip')
        ])
    ]

    latest_cache_file = sorted(
        cache_files,
        key=lambda x: version.parse(os.path.splitext(x.split('_')[-1])[0]),
        reverse=True
    )[0]

    shutil.rmtree(CACHE_DIR, ignore_errors=True)

    latest_cache_filepath = os.path.join(BASE_DIR, latest_cache_file)
    with zipfile.ZipFile(latest_cache_filepath, 'r') as zip_ref:
        zip_ref.extractall(BASE_DIR)

    # make sure that the version numbers of whatever cache we downloaded
    # are changed to the current version so that the caches can be picked up
    for root, dirs, files in os.walk(CACHE_DIR):
        for dir_name in dirs:
            original = os.path.join(root, dir_name)
            new_dir = re.sub(r'/\d+\.\d+\.\d+/?', f'/{VERSION}/', original)
            shutil.move(original, new_dir)


def is_cache_already_pulled() -> bool:
    """
    Determine if there is already a valid cache to use

    Could be smarter about this and see if the cache is somewhat full
    """

    already_pulled = all([
        os.path.exists(os.path.join(EDGE_CACHE_DIR, VERSION)),
        os.path.exists(os.path.join(NETWORK_CACHE_DIR, VERSION)),
        os.path.exists(os.path.join(BOUNDING_GRAPH_GDFS_CACHE, VERSION)),
    ])
    if already_pulled:
        logger.info('Assuming the cache is up to date')
    else:
        logger.warning('Cache is not up to date')
    return already_pulled



import os
import shutil
import glob

import boto3
import requests
from botocore import UNSIGNED
from botocore.client import Config

from via import logger
from via.settings import (
    DOWNLOAD_JOURNEYS_URL,
    S3_REGION
)
from via.constants import REMOTE_DATA_DIR
from via.models.journey import Journey


def pull_journeys(cache_graphs=False):
    """

    :kwargs cache_graphs: cache the graphs of pulled journeys so we don't
        need to do it again later when we need them
    """
    s3 = boto3.client(
        's3',
        region_name=S3_REGION,
        config=Config(signature_version=UNSIGNED)
    )

    if not os.path.exists(REMOTE_DATA_DIR):
        os.makedirs(REMOTE_DATA_DIR, exist_ok=True)

    journey_ids = []
    for filename in glob.iglob(REMOTE_DATA_DIR + '/**/*', recursive=True):
        journey_ids.append(os.path.splitext(os.path.basename(filename))[0])

    journey_filenames = requests.get(DOWNLOAD_JOURNEYS_URL).json()
    journey_files_to_download = []

    for filename in journey_filenames:
        journey_id = os.path.splitext(os.path.basename(filename))[0]
        if journey_id in journey_ids:
            continue
        journey_files_to_download.append(filename)

    logger.info(f'Downloading {len(journey_files_to_download)} files')

    for filename in journey_files_to_download:

        journey_id = os.path.splitext(filename)[0]

        tmp_filepath = f'/tmp/{journey_id}.json'

        s3.download_file(
            'bike-road-quality',
            filename,
            tmp_filepath
        )

        journey = Journey.from_file(tmp_filepath)

        geo = journey.origin.gps.reverse_geo
        logger.info(f'Pulled journey from: {geo["cc"]}, {geo["place_3"]}, {geo["place_2"]}, {geo["place_1"]}')

        if cache_graphs:
            logger.info(f'Caching graphs for {journey_id}')
            journey.bounding_graph

        local_filepath = os.path.join(
            REMOTE_DATA_DIR,
            journey.transport_type.lower(),
            filename
        )
        logger.info(f'Putting to {local_filepath}')

        os.makedirs(
            os.path.dirname(local_filepath),
            exist_ok=True
        )

        shutil.move(
            tmp_filepath,
            local_filepath
        )



import osmnx as ox


class NearestNodeCache():

    def __init__(self):
        self.data = {}

    def get(self, graph, frames, return_dist=False):
        """

        :param graph:
        :param frames: list of frames to get nearest nodes of
        :kwarg return_dist: to return the distance alongside the node id
        :rtype: list or list of tuples
        :return: list of of node ids or tuple of node_id and distance
        """
        frame_ids_to_get = [
            frame.uuid for frame in frames if self.data.get(frame.uuid, None) is None
        ]

        id_frame_map = {f.uuid: f for f in frames}

        requested_frame_node_map = {
            f.uuid: self.data.get(f.uuid, None) for f in frames
        }

        if frame_ids_to_get != []:
            results = ox.distance.nearest_nodes(
                graph,
                [id_frame_map[frame_id].gps.lng for frame_id in frame_ids_to_get],
                [id_frame_map[frame_id].gps.lat for frame_id in frame_ids_to_get],
                return_dist=True
            )

            frame_id_result_map = dict(
                zip(
                    frame_ids_to_get,
                    list(
                        zip(
                            results[0],
                            results[1]
                        )
                    )
                )
            )

            for frame_id, node_data in frame_id_result_map.items():
                self.data[frame_id] = node_data

        requested_frame_node_map = {
            f.uuid: self.data.get(f.uuid, None) for f in frames
        }

        if return_dist:
            return requested_frame_node_map.values()
        return [v[0] for v in requested_frame_node_map.values()]


nearest_node = NearestNodeCache()



# used for finding place bounds, wuld take too long to actually get
# graphs from osmnx

from typing import Mapping

import osmnx as ox

from via.utils import is_within


class PlaceCache():

    def __init__(self):
        self.data = {
            'dublin ireland': {
                'north': 53.626487,
                'south': 53.3018049,
                'east': -6.1366563,
                'west': -6.433500
            }
        }

    def get_by_bbox(self, cmp_bbox):
        for name, bbox in self.data.items():
            if is_within(cmp_bbox, bbox):
                return {
                    'name': name,
                    'bbox': bbox
                }
        return None

    def is_in_place(self, bbox: dict, place_name: str):
        place_name = place_name.lower().replace(',', '')
        try:
            return is_within(bbox, self.data[place_name])
        except KeyError:
            return False

    def get(self, place_name: str) -> Mapping[str, float]:
        """
        Get the lat long bounds of a place

        :param place_name: An osmnx recognised place name
            (for example "Dublin, Ireland")
        :return: A dict of north, south, east, and west
            lats and lngs
        """
        place_name = place_name.lower().replace(',', '')
        try:
            return self.data[place_name]
        except KeyError:
            place_graph = ox.graph_from_place(
                place_name,
                network_type='all'
            )

            lats = []
            lngs = []
            for _, node_data in place_graph._node.items():
                lats.append(node_data['y'])
                lngs.append(node_data['x'])

            self.data[place_name] = {
                'north': max(lats),
                'south': min(lats),
                'east': max(lngs),
                'west': min(lngs)
            }
            return self.data[place_name]


place_cache = PlaceCache()



from cached_property import cached_property

from typing import Tuple

import reverse_geocoder as rg
from haversine import (
    haversine,
    Unit
)

HAVERSINE_CACHE = {}


class GPSPoint():
    """
    GPSPoint is an object containing geospatial data in three directions.

    Should be used instead of tuples of (lat, lng) or whatever since
    sometimes libraries expect (lng, lat)
    """

    def __del__(self):
        attrs_to_del = [
            'reverse_geo',
            'content_hash'
        ]

        for attr in attrs_to_del:
            try:
                delattr(self, attr)
            except:
                pass

    def __init__(self, lat: float, lng: float, elevation=None):
        """

        :param lat:
        :param lng:
        :kwarg elevation: Optional elevation in metres
        """
        self.lat = lat
        self.lng = lng
        self.elevation = elevation

    def __eq__(self, oth):
        return self.lat == oth.lat and self.lng == oth.lng

    @staticmethod
    def parse(obj):
        if isinstance(obj, list):
            return GPSPoint(
                obj[0],
                obj[1]
            )

        if isinstance(obj, GPSPoint):
            return obj

        if isinstance(obj, dict):
            return GPSPoint(
                obj['lat'],
                obj['lng'],
                elevation=obj.get('elevation', None)
            )

        raise NotImplementedError(
            f'Can\'t parse gps from type {type(obj)}'
        )

    def distance_from(self, point) -> float:
        """

        :param point: GPSPoint or tuple of (lat, lng)
        :rtype: float
        :return: Distance between points in metres
        """
        if isinstance(point, GPSPoint):
            point = point.point

        key = hash((self.point, point))
        if key not in HAVERSINE_CACHE:
            HAVERSINE_CACHE[key] = haversine(
                self.point,
                point,
                unit=Unit.METERS
            )

        return HAVERSINE_CACHE[key]

    def slope_between(self, dst):
        from via.models.point import FramePoint
        if isinstance(dst, FramePoint):
            dst = dst.gps
        try:
            return (self.lng - dst.lng) / (self.lat - dst.lat)
        except ZeroDivisionError:
            return 0

    def serialize(self):
        return {
            'lat': self.lat,
            'lng': self.lng,
            'elevation': self.elevation
        }

    @cached_property
    def reverse_geo(self):
        # TODO: make a cache for this for "close enough" positions if we end
        # up using this frequently
        data = dict(rg.search(
            (self.lat, self.lng)
        )[0])
        del data['lat']
        del data['lon']
        data['place_1'] = data.pop('name', None)
        data['place_2'] = data.pop('admin1', None)
        data['place_3'] = data.pop('admin2', None)
        return data

    @cached_property
    def content_hash(self) -> str:
        """
        A content hash that will act as an id for the data, handy for caching
        """
        return int.from_bytes(f'{self.lat} {self.lng} {self.elevation}'.encode(), 'little') % 2**100

    @property
    def point(self) -> Tuple[float]:
        """

        :rtype: tuple
        :return: tuple of (lat, lng)
        """
        return (self.lat, self.lng)

    @property
    def is_populated(self) -> bool:
        """
        Is the gps data populated. Often when there is no satelite
        or starting up this will not be populated.

        Does not consider elevation since it's not important

        :rtype: bool
        """
        return isinstance(self.lat, (int, float)) and isinstance(self.lng, (int, float)) and self.lat != 0 and self.lng != 0



import datetime
import os
import urllib
from packaging.version import Version

import boto3
from botocore import UNSIGNED
from botocore.client import Config

import shapely
from networkx.readwrite import json_graph

from via.settings import (
    S3_REGION,
    PREPARED_GEOJSON_BUCKET
)
from via.constants import GEOJSON_DIR


def parse_start_date(earliest_date):
    if earliest_date is None:
        return '2021-01'

    if earliest_date < datetime.datetime(2021, 1, 1):
        return '2021-01'

    return earliest_date


def parse_end_date(latest_date):
    if latest_date is None:
        return '2023-12'

    if latest_date > datetime.datetime(2023, 12, 31):
        return '2023-12'

    return latest_date


def generate_basename(
    name=None,
    version=None,
    version_op=None,
    earliest_time=None,
    latest_time=None,
    place=None
):
    data = {
        'transport_type': name,
        'version': version,
        'version_op': version_op,
        'earliest_time': parse_start_date(earliest_time),
        'latest_time': parse_end_date(latest_time),
        'place': place
    }
    if data['version'] in {None, False, '0.0.0', Version('0.0.0')}:
        del data['version']
        del data['version_op']
    if data['place'] is None:
        del data['place']
    return urllib.parse.urlencode(data)


def geojson_from_graph(graph):
    json_links = json_graph.node_link_data(
        graph
    )['links']

    geojson_features = {
        'type': 'FeatureCollection',
        'features': []
    }

    for link in json_links:
        if 'geometry' not in link:
            continue

        feature = {
            'type': 'Feature',
            'properties': {}
        }

        for k in link:
            if k == 'geometry':
                feature['geometry'] = shapely.geometry.mapping(
                    link['geometry']
                )
            else:
                feature['properties'][k] = link[k]
        useless_properties = {
            'oneway',
            'length',
            'osmid',
            'highway',
            'source',
            'target',
            'key',
            'maxspeed',
            'lanes',
            'ref'
        }
        for useless_property in useless_properties:
            try:
                del feature['properties'][useless_property]
            except:
                pass
        geojson_features['features'].append(feature)

    return geojson_features


def download_prepared_geojson():
    os.makedirs(
        GEOJSON_DIR,
        exist_ok=True
    )

    s3 = boto3.client(
        's3',
        region_name=S3_REGION,
        config=Config(signature_version=UNSIGNED)
    )

    paginator = s3.get_paginator('list_objects')
    for result in paginator.paginate(Bucket=PREPARED_GEOJSON_BUCKET):
        for key in result.get('Contents', []):
            s3.download_file(
                PREPARED_GEOJSON_BUCKET,
                key['Key'],
                os.path.join(GEOJSON_DIR, key['Key'])
            )
            os.utime(
                os.path.join(
                    GEOJSON_DIR,
                    key['Key']
                ),
                (
                    key['LastModified'].timestamp(),
                    key['LastModified'].timestamp()
                )
            )



import json
import pickle
import random

from mock import patch

import osmnx as ox

from unittest import TestCase

from via.nearest_node import NearestNodeCache
from via.models.frame import Frame, Frames


class NearestNodeTest(TestCase):

    def setUp(self):
        with open('test/resources/dublin_graph.pickle', 'rb') as f:
            self.graph = pickle.load(f)
        self.points = [
            (-6.295029, 53.348127),
            (-6.259481, 53.344463),
            (-6.260696, 53.339891),
            (-6.258941, 53.346979)
        ]

    def test_get_nothing_cached(self):
        cache = NearestNodeCache()
        frames = []
        for point in self.points:
            frames.append(
                Frame(
                    0,
                    {'lat': point[1], 'lng': point[0]},
                    (0, 0, 0)
                )
            )

        self.assertEqual(
            cache.get(self.graph, frames),
            [12428414, 3451879900, 2893792091, 5101079533]
        )

    def test_get_some_cached(self):
        cache = NearestNodeCache()

        frames = []
        for point in random.sample(self.points, int(len(self.points) / 2)):
            frames.append(
                Frame(
                    0,
                    {'lat': point[1], 'lng': point[0]},
                    (0, 0, 0)
                )
            )
        cache.get(self.graph, frames)

        frames = []
        for point in self.points:
            frames.append(
                Frame(
                    0,
                    {'lat': point[1], 'lng': point[0]},
                    (0, 0, 0)
                )
            )
        self.assertEqual(
            cache.get(self.graph, frames),
            [12428414, 3451879900, 2893792091, 5101079533]
        )

    def test_get_all_cached(self):
        cache = NearestNodeCache()

        frames = []
        for point in self.points:
            frames.append(
                Frame(
                    0,
                    {'lat': point[1], 'lng': point[0]},
                    (0, 0, 0)
                )
            )
        self.assertEqual(
            cache.get(self.graph, frames),
            [12428414, 3451879900, 2893792091, 5101079533]
        )
        self.assertEqual(
            cache.get(self.graph, frames),
            [12428414, 3451879900, 2893792091, 5101079533]
        )



import pickle

from unittest import TestCase

from via.nearest_edge import NearestEdgeCache
from via.models.point import FramePoint


class NearestEdgeTest(TestCase):

    def setUp(self):
        with open('test/resources/dublin_graph.pickle', 'rb') as f:
            self.graph = pickle.load(f)

    def test_get(self):
        cache = NearestEdgeCache()
        edges = cache.get(
            self.graph,
            [
                FramePoint.parse(
                    {
                        'time': 0,
                        'gps': {'lat': 53.348127, 'lng': -6.295029},
                        'acc': [0, 0, 0]
                    }
                )
            ]
        )

        print(f'Edges pre: {edges}')
        edges = [list(edges[0][0]) for i in edges]
        print(f'Edges post: {edges}')
        self.assertTrue(
            (12428414, 4161475238, 0) in edges[0] or (4161475238, 12428414, 0) in edges[0] or [12428414, 4161475238, 0] in edges[0] or [4161475238, 12428414, 0] in edges[0]
        )

    def test_save_load(self):
        cache = NearestEdgeCache()
        cache.get(
            self.graph,
            [
                FramePoint.parse(
                    {
                        'time': 0,
                        'gps': {'lat': 53.348127, 'lng': -6.295029},
                        'acc': [0, 0, 0]
                    }
                )
            ]
        )
        cache = NearestEdgeCache()
        cache.load()
        print(cache.data.keys())
        self.assertTrue('489334746920999497411653940021' in cache.data.keys())



import hashlib
import urllib
import os
import json
import glob
from multiprocessing import (
    Pool,
    cpu_count
)
from multiprocessing.dummy import Pool as ThreadPool
from itertools import groupby
from operator import itemgetter
from collections import defaultdict

from cached_property import cached_property
from haversine import (
    haversine,
    Unit
)

import osmnx as ox

from road_collisions.models.raw_collision import RawCollision as BaseRawCollision
from road_collisions.models.collision import Collision as BaseCollision
from road_collisions.models.collision import Collisions as BaseCollisions

from via import logger
from via.constants import GEOJSON_DIR
from via.nearest_edge import nearest_edge
from via.constants import METRES_PER_DEGREE
from via.utils import (
    area_from_coords,
    get_graph_id,
    filter_nodes_from_geodataframe,
    filter_edges_from_geodataframe,
    update_edge_data,
    get_combined_id
)
from via.network_cache import network_cache
from via.models.gps import GPSPoint
from via.bounding_graph_gdfs_cache import bounding_graph_gdfs_cache
from via.geojson.utils import geojson_from_graph


class Collision(BaseCollision):

    def __del__(self):
        attrs_to_del = [
            'gps_hash',
            'lat_span',
            'lng_span',
            'upper_left_bbox',
            'upper_right_bbox',
            'lower_left_bbox',
            'lower_right_bbox'
        ]

        for attr in attrs_to_del:
            try:
                delattr(self, attr)
            except:
                pass

    @staticmethod
    def parse(data):
        if isinstance(data, Collision):
            return data

        if isinstance(data, dict):
            if 'data' in data.keys():
                return Collision(
                    **BaseRawCollision.parse(
                        data
                    ).data
                )
            else:
                # from serialization
                return Collision(
                    **data
                )

    @property
    def gps(self):
        return GPSPoint(
            self.lat,
            self.lng
        )

    @cached_property
    def gps_hash(self):
        return self.gps.content_hash

    def is_in_place(self, place_bounds):
        return all([
            self.lat < place_bounds['north'],
            self.lat > place_bounds['south'],
            self.lng < place_bounds['east'],
            self.lng > place_bounds['west']
        ])


class Collisions(BaseCollisions):

    def __init__(self, *args, **kwargs):
        """

        :kwarg filters: Filters to apply, these filters will apply inplace
            as a relevant action is made
        """
        self.filters = kwargs.get('filters', {})
        self.is_filtered = kwargs.get('is_filtered', False)
        super().__init__(*args, **kwargs)

    def set_filters(self, filters):
        self.filters = filters

    @property
    def danger(self):
        # FIXME: generally improve
        return len(self)

    @property
    def danger_by_vehicle_type(self):

        type_collisions = [
            (c.vehicle_type, c) for c in self
        ]

        d = defaultdict(Collisions)
        for x, y in type_collisions:
            d[x].append(y)

        return {
            k: v.danger for k, v in d.items()
        }

    @property
    def most_northern(self):
        return max([collision.gps.lat for collision in self])

    @property
    def most_southern(self):
        return min([collision.gps.lat for collision in self])

    @property
    def most_eastern(self):
        return max([collision.gps.lng for collision in self])

    @property
    def most_western(self):
        return min([collision.gps.lng for collision in self])

    @property
    def bbox(self):
        return {
            'north': self.most_northern,
            'south': self.most_southern,
            'east': self.most_eastern,
            'west': self.most_western
        }

    @cached_property
    def lat_span(self):
        return haversine(
            (self.most_northern, 0),
            (self.most_southern, 0),
            unit=Unit.METERS
        )

    @cached_property
    def lng_span(self):
        return haversine(
            (0, self.most_western),
            (0, self.most_eastern),
            unit=Unit.METERS
        )

    @cached_property
    def upper_left_bbox(self):
        return {
            'north': self.most_northern,
            'south': self.most_southern + ((self.lat_span / METRES_PER_DEGREE) / 2),
            'west': self.most_western,
            'east': self.most_western + ((self.lng_span / METRES_PER_DEGREE) / 2),
        }

    @property
    def upper_left_collisions(self):
        return Collisions(
            data=[
                collision for collision in self if collision.is_in_place(self.upper_left_bbox)
            ]
        )

    @cached_property
    def upper_right_bbox(self):
        return {
            'north': self.most_northern,
            'south': self.most_southern + ((self.lat_span / METRES_PER_DEGREE) / 2),
            'west': self.most_western + ((self.lng_span / METRES_PER_DEGREE) / 2),
            'east': self.most_eastern,
        }

    @property
    def upper_right_collisions(self):
        return Collisions(
            data=[
                collision for collision in self if collision.is_in_place(self.upper_right_bbox)
            ]
        )

    @cached_property
    def lower_left_bbox(self):
        return {
            'north': self.most_southern + ((self.lat_span / METRES_PER_DEGREE) / 2),
            'south': self.most_southern,
            'west': self.most_western,
            'east': self.most_western + ((self.lng_span / METRES_PER_DEGREE) / 2),
        }

    @property
    def lower_left_collisions(self):
        return Collisions(
            data=[
                collision for collision in self if collision.is_in_place(self.lower_left_bbox)
            ]
        )

    @cached_property
    def lower_right_bbox(self):
        return {
            'north': self.most_southern + ((self.lat_span / METRES_PER_DEGREE) / 2),
            'south': self.most_southern,
            'west': self.most_western + ((self.lng_span / METRES_PER_DEGREE) / 2),
            'east': self.most_eastern,
        }

    @property
    def lower_right_collisions(self):
        return Collisions(
            data=[
                collision for collision in self if collision.is_in_place(self.lower_right_bbox)
            ]
        )

    def preload_graph(self, use_graph_cache=True):

        if use_graph_cache is False or network_cache.get('bbox', self) is None:

            if area_from_coords(self.bbox) > 1000000000:
                logger.debug('Graph too large to load, splitting: %s', self.bbox)

                if len(self.upper_left_collisions) > 1:
                    self.upper_left_collisions.preload_graph()

                if len(self.upper_right_collisions) > 1:
                    self.upper_right_collisions.preload_graph()

                if len(self.lower_left_collisions) > 1:
                    self.lower_left_collisions.preload_graph()

                if len(self.lower_right_collisions) > 1:
                    self.lower_right_collisions.preload_graph()

            else:
                logger.debug('No split necessary in loading graph: %s', self.bbox)
                try:
                    # TODO: depending on filters, might want to get different network. Only for geojson generation. If filter of vehicle_type being car / goods vehicle, use car, if bike / pedestrian, use all
                    #       will need to have the network_cache accept network types to split then
                    network = ox.graph_from_bbox(
                        self.most_northern,
                        self.most_southern,
                        self.most_eastern,
                        self.most_western,
                        network_type='all',
                        simplify=True
                    )
                    network_cache.set(
                        'bbox',
                        self,
                        network
                    )
                except Exception as ex:
                    logger.warning('Could not create graph %s: %s', self.bbox, ex)

    @staticmethod
    def from_file(filepath):
        data = None
        with open(filepath, 'r') as collisions_filepath:
            data = json.loads(collisions_filepath.read())

        collisions = Collisions()
        for collision_dict in data:
            collisions.append(
                Collision.parse(
                    collision_dict
                )
            )

        return collisions

    @staticmethod
    def from_dir(dirpath):
        collisions = Collisions()
        for filename in glob.iglob(f'{dirpath}/**', recursive=True):
            if os.path.splitext(filename)[-1] != '.json':
                continue
            collisions.extend(
                Collisions.from_file(
                    filename
                )._data
            )

        return collisions

    @staticmethod
    def load_all():
        import road_collisions
        return Collisions.from_dir(
            os.path.join(road_collisions.__path__[0], 'resources')
        )

    def filter(self, **kwargs):
        '''
        By whatever props that exist
        '''
        logger.debug('Filtering from %s' % (len(self)))

        filtered = []
        for collision in self:
            valid = True
            for attr, val in kwargs.items():
                if getattr(collision, attr) != val:
                    valid = False
                    break
            if valid:
                filtered.append(collision)

        return Collisions(
            data=filtered,
            filters=kwargs,
            is_filtered=True
        )

    def inplace_filter(self, **kwargs):
        self.filters = kwargs
        if self.filters != {}:
            self._data = self.filter(**kwargs)._data
        self.is_filtered = True

    @property
    def gps_hash(self):
        if not self.is_filtered:
            self.inplace_filter(**self.filters)

        return hashlib.md5(
            str([
                point.gps.content_hash for point in self
            ]).encode()
        ).hexdigest()

    @staticmethod
    def split_collisions(collision):
        """
        Get the associated graph id and the give collision
        This can be then used to group collisions by graph

        :rtype: tuple
        """
        network = network_cache.get_at_point('bbox', collision)
        if network is not None:
            # this is slowing things down a lot, have a cache
            return get_graph_id(network), collision

        return None, None

    @property
    def network_collision_map(self):

        def get_mapping(graph_id_collisions):
            graph_id = graph_id_collisions[0]
            collisions = graph_id_collisions[1]
            return {
                'network': network_cache.get_by_id(graph_id),
                'collisions': collisions,
                'graph_id': graph_id
            }

        if not self.is_filtered:
            self.inplace_filter(**self.filters)

        maps = [Collisions.split_collisions(c) for c in self]
        maps = [m for m in maps if m[0] is not None]

        network_id_collisions_map = [
            (k, Collisions(data=[x for _, x in g])) for k, g in groupby(
                sorted([i for i in maps if i is not None], key=itemgetter(0)),
                itemgetter(0)
            )
        ]

        mapping_pool = ThreadPool(5)
        results = mapping_pool.map(get_mapping, network_id_collisions_map)
        mapping_pool.close()
        mapping_pool.join()

        network_collision_map = {}
        for map_data in results:
            network_collision_map[map_data['graph_id']] = {
                'network': map_data['network'],
                'collisions': map_data['collisions']
            }

        return network_collision_map

    @property
    def fp(self):
        filters = {k: v for k, v in self.filters.items() if v is not None}
        return os.path.join(
            GEOJSON_DIR,
            'collision_' + urllib.parse.urlencode(filters) + '.geojson'
        )

    @property
    def percent_collisions_outside_networks(self):
        """
        Return the % of collisions that are not within already cached networks
        """
        # Before preloading graph, see if we can get pretty much all points
        bad = 0
        good = 0
        for collision in self:
            at_point = network_cache.get_at_point('bbox', collision)
            if at_point is None:
                bad += 1
            else:
                good += 1

        if good == 0:
            if bad == 0:
                perc_bad = 0
            else:
                perc_bad = 100
        else:
            perc_bad = (bad / (good + bad)) * 100
            logger.debug('Missing %s%% of collisions', perc_bad)

        return perc_bad

    # want to get {edge_id: [collision, collision]}

    @property
    def edge_collision_map(self):

        data = defaultdict(Collisions)

        for _, network_collision_dict in self.network_collision_map.items():

            network = network_collision_dict['network']
            collisions = network_collision_dict['collisions']

            edges = nearest_edge.get(network, collisions)

            used_edges = []
            used_node_ids = []
            for edge in edges:
                # TODO: get closest, we just use the first
                used_edges.append(tuple(edge[0][0]))
                used_node_ids.extend(
                    [
                        edge[0][0][0],
                        edge[0][0][1]
                    ]
                )

            associated = list(zip(used_edges, collisions))

            for k, g in groupby(
                sorted(associated, key=itemgetter(0)),
                itemgetter(0)
            ):
                # TODO: are there ever overwrites?
                data[get_combined_id(k[0], k[1])] = Collisions(
                    data=[x for _, x in g]
                )

        return data

    @property
    def geojson(self):
        if os.path.exists(self.fp):
            with open(self.fp, 'r') as geojson_file:
                return json.loads(geojson_file.read())

        if not self.is_filtered:
            self.inplace_filter(**self.filters)

        if self.percent_collisions_outside_networks > 2:
            logger.debug('Not all collision networks present, need to preload')

            self.preload_graph()

        geojson_features = []

        for _, network_collision_dict in self.network_collision_map.items():

            network = network_collision_dict['network']
            collisions = network_collision_dict['collisions']

            edges = nearest_edge.get(network, collisions)

            used_edges = []
            used_node_ids = []
            for edge in edges:
                # TODO: get closest, we just use the first
                used_edges.append(tuple(edge[0][0]))
                used_node_ids.extend(
                    [
                        edge[0][0][0],
                        edge[0][0][1]
                    ]
                )

            associated = list(zip(used_edges, collisions))
            edge_feature_map = {
                get_combined_id(k[0], k[1]): {
                    'danger': Collisions(
                        data=[x for _, x in g]
                    ).danger
                } for k, g in groupby(
                    sorted(associated, key=itemgetter(0)),
                    itemgetter(0)
                )
            }

            if bounding_graph_gdfs_cache.get(get_graph_id(network)) is None:
                gdfs_graph = ox.graph_to_gdfs(
                    network,
                    fill_edge_geometry=True
                )
                bounding_graph_gdfs_cache.set(
                    get_graph_id(network),
                    gdfs_graph
                )

            graph_nodes, graph_edges = bounding_graph_gdfs_cache.get(
                get_graph_id(network)
            )

            graph = ox.graph_from_gdfs(
                filter_nodes_from_geodataframe(graph_nodes, used_node_ids),
                filter_edges_from_geodataframe(graph_edges, used_edges)
            )

            update_edge_data(graph, edge_feature_map)

            features = geojson_from_graph(graph)['features']

            geojson_features.extend(features)

        geojson = {
            'type': 'FeatureCollection',
            'features': geojson_features
        }

        with open(self.fp, 'w') as geojson_file:
            geojson_file.write(json.dumps(geojson))

        return geojson



from via.utils import get_combined_id


EDGE_CACHE = {}


def get_edge_data(origin_uuid: str, destination_uuid: str, graph=None):
    """

    :param origin_uuid: osmid of the origin node
    :param destination_uuid: osmid of the destination node
    :return: TODO: can't remember exactly what gets returned
    """

    combined_id = get_combined_id(origin_uuid, destination_uuid)

    try:
        return EDGE_CACHE[combined_id]
    except KeyError:
        if graph is None:
            # Since we can provide no more context
            return None

        EDGE_CACHE[combined_id] = graph.get_edge_data(
            origin_uuid,
            destination_uuid
        )
        return EDGE_CACHE[combined_id]



import os
import time
import stat

import fast_json

from via.constants import GEOJSON_DIR
from via.geojson.utils import generate_basename


def get_geojson(
    journey_type,
    earliest_time=None,
    latest_time=None,
    place=None,
    version=None,
    version_op=None,
    max_age=60*60*12
):
    if journey_type is None:
        journey_type = 'all'

    basename = generate_basename(
        name=journey_type,
        version=version,
        version_op=version_op,
        earliest_time=earliest_time,
        latest_time=latest_time,
        place=place
    )
    geojson_file = os.path.join(
        GEOJSON_DIR,
        f'{basename}.geojson'
    )

    if not os.path.exists(geojson_file):
        raise FileNotFoundError()

    if time.time() - os.stat(geojson_file)[stat.ST_MTIME] > max_age:
        raise FileNotFoundError()

    geojson_data = []
    with open(geojson_file) as fh:
        geojson_data = fast_json.load(fh)

    return geojson_data



import os
import logging
import threading
import datetime
import pickle
import json
from typing import List, Any

from via import logger
from via import settings
from via.constants import CACHE_DIR
from via.settings import CLEAN_MEMORY
from via.utils import get_size


class BaseCache():

    def __init__(self, cache_type: str = None, fn: str = None):
        assert cache_type is not None
        self.loaded = False
        self.data = {}
        self.last_save_len = -1
        self.cache_type = cache_type
        self.fn = fn
        self.last_accessed = datetime.datetime.utcnow()

    def get(self, k: Any):
        self.load()
        self.last_accessed = datetime.datetime.utcnow()
        return self.data.get(k, None)

    def set(self, k, v, skip_save=False):
        self.load()
        self.data[k] = v
        if not skip_save:
            self.save()
        self.last_accessed = datetime.datetime.utcnow()

    def create_dirs(self):
        if not os.path.exists(self.fp):
            logger.debug(f'Creating cache dir: {self.fp}')
            os.makedirs(
                os.path.dirname(self.fp),
                exist_ok=True
            )

    def save(self):
        if len(self.data) <= self.last_save_len:
            return
        logger.info(f'Saving cache {self.fp}')
        self.create_dirs()
        with open(self.fp, 'wb') as f:
            pickle.dump(self.data, f)
        self.last_save_len = len(self.data)

    def load(self):
        if self.loaded:
            return

        logger.info(f'Loading cache {self.fp}')
        if not os.path.exists(self.fp):
            self.create_dirs()
            self.save()

        with open(self.fp, 'rb') as f:
            self.data = pickle.load(f)

        self.loaded = True
        self.last_save_len = len(self.data)

        if logger.level <= logging.DEBUG:
            logger.debug(
                'Size of %s cache: %sMB',
                self.fp,
                get_size(self.data) / (1000 ** 2)
            )

    def unload(self):
        self.data = {}
        self.loaded = False

    @staticmethod
    def from_file(cache_type: str, filepath: str, load: bool = True, fn: str = None):
        cache = BaseCache(
            cache_type=cache_type,
            fn=os.path.basename(filepath)
        )

        if load:
            cache.load()

        return cache

    @property
    def dir(self) -> str:
        return os.path.join(CACHE_DIR, self.cache_type, settings.VERSION)

    @property
    def fp(self) -> str:
        return os.path.join(self.dir, f'{self.fn}')

    @property
    def since_last_accessed(self):
        return (datetime.datetime.utcnow() - self.last_accessed).total_seconds()


class BaseCaches():

    def __init__(self, *args, **kwargs):
        assert kwargs.get('cache_type', None) is not None
        self.loaded = False
        self.data = {}
        self.last_save_len = -1
        self.cache_type = kwargs['cache_type']
        self.child_class = kwargs.get('child_class', BaseCache)

        # set up references
        self.refs = {}
        os.makedirs(self.dir, exist_ok=True)
        if not os.path.exists(self.refs_path):
            with open(self.refs_path, 'w') as refs_file:
                refs_file.write(json.dumps({}))
        with open(self.refs_path, 'r') as refs_file:
            self.refs = json.loads(refs_file.read())

        if CLEAN_MEMORY:
            self.memory_cleaner()

    def get(self, k: Any):
        self.load()
        if k in self.refs and self.refs[k] in self.data:
            return self.data[self.refs[k]].get(k)
        return None

    def get_fn(self, obj):
        raise NotImplementedError()

    def set(self, k: Any, v: Any, skip_save: bool = False):
        fn = self.get_fn(v)

        if fn not in self.data:
            self.data[fn] = self.child_class(fn=fn)
        self.data[fn].set(k, v)
        len_refs = len(self.refs)
        self.refs[k] = fn
        if len_refs < len(self.refs):
            if not skip_save:
                self.data[fn].save()
                self.save_refs()

    def save_refs(self):
        with open(self.refs_path, 'w') as refs_file:
            refs_file.write(json.dumps(self.refs))

    def load(self):
        """
        Load all the caches of this type, note that this does not load
        data, it loads only references so that things can be lazily loaded
        """
        if self.loaded:
            return

        try:
            for c in self.caches:
                self.data[c.fn] = c
        except:
            self.data = {}

        self.loaded = True

    def memory_cleaner(self):
        """
        Clean up memory from caches that haven't been used in a while
        """
        logger.debug('Cleaning memory: %s', self.cache_type)

        if logger.level <= logging.DEBUG:
            # since get_size is a little slow, don't want to waste time if
            # we're not going to log about it
            try:
                initial_memory = get_size(self)
            except:
                initial_memory = -1

        for v in self.data.values():
            if v.since_last_accessed > 60:
                v.unload()

        if logger.level <= logging.DEBUG:
            try:
                post_memory = get_size(self)
            except:
                post_memory = -1
            logger.debug(
                'Cleaned memory, reduced %s by %s%% (%s -> %s)',
                self.cache_type,
                ((float(post_memory) - initial_memory) / initial_memory) * 100,
                initial_memory,
                post_memory
            )

        cleaner = threading.Timer(
            60 * 2,
            self.memory_cleaner
        )
        cleaner.daemon = True
        cleaner.start()

    @property
    def dir(self) -> str:
        return os.path.join(CACHE_DIR, self.cache_type, settings.VERSION)

    @property
    def caches(self) -> List[BaseCache]:
        """
        Get all the caches within the directory of this cache type
        """
        return [
            self.child_class.from_file(
                self.cache_type,
                os.path.join(self.dir, f),
                load=False
            ) for f in os.listdir(self.dir)
        ]

    @property
    def refs_path(self) -> str:
        """
        Get the reference file path for this type of cache
        """
        return os.path.join(self.dir, 'refs.json')



import json
import datetime
import logging
import os
import pickle
import threading

import osmnx as ox
from networkx.classes.multidigraph import MultiDiGraph

from via import settings
from via.constants import NETWORK_CACHE_DIR, CACHE_DIR
from via import logger
from via.utils import (
    is_within,
    area_from_coords,
    get_graph_id,
    get_size
)
from via.place_cache import place_cache


class SingleNetworkCache():

    def __init__(self, network_type: str, fn=None):
        self.fn = fn
        self.network_type = network_type
        self.loaded = False
        self.networks_loaded = False
        self.data = {}
        self.networks = {}
        self.last_save_len = -1
        self.last_accessed = datetime.datetime.utcnow()

    @staticmethod
    def from_file(cache_type: str, filepath: str, load: bool = True, fn: str = None):
        cache = SingleNetworkCache(
            network_type=cache_type,
            fn=os.path.basename(filepath)
        )

        if load:
            cache.load()

        return cache

    def get_by_id(self, graph_id):
        if graph_id not in self.data:
            return None
        self.load_networks(network_id=graph_id)
        return {
            **self.data[graph_id],
            'network': self.networks[graph_id]
        }

    def get_at_point(self, gps_point):
        self.last_accessed = datetime.datetime.utcnow()

        self.load()

        point_bbox = {
            'north': gps_point.lat,
            'south': gps_point.lat,
            'east': gps_point.lng,
            'west': gps_point.lng
        }
        candidates = []
        for k, net in self.data.items():
            if is_within(point_bbox, net['bbox']):
                candidates.append(k)
                break

        if not len(candidates):
            return None

        self.load_networks(network_id=candidates[0])
        # FIXME: return the smallest network
        return self.networks[candidates[0]]

    def load_networks(self, network_id=None):
        if self.networks_loaded:
            return

        if network_id in self.networks:
            return

        os.makedirs(
            os.path.join(self.dir, 'networks'),
            exist_ok=True
        )
        network_files = os.listdir(os.path.join(self.dir, 'networks'))

        for network_filename in network_files:
            if os.path.splitext(network_filename)[0] in self.data.keys():
                network_filepath = os.path.join(self.dir, 'networks', network_filename)
                graph_id = os.path.splitext(os.path.basename(network_filepath))[0]
                if graph_id in self.networks:
                    # already loaded from a specific network_id previously
                    continue
                if network_id is not None:
                    if graph_id != network_id:
                        continue
                with open(network_filepath, 'rb') as network_file:
                    self.networks[graph_id] = pickle.load(network_file)

        if network_id is None:
            # It's only fully loaded when no network specified
            self.networks_loaded = True

    def get(self, journey) -> MultiDiGraph:
        self.last_accessed = datetime.datetime.utcnow()

        self.load()

        # if not poly and we find within a bbox we should be able to do that
        # truncate_graph_polygon. Don't need to save that poly graph if we
        # want to optimize for storage. Can see how it does anyways

        if self.network_type != 'poly':
            candidates = []
            for k, net in self.data.items():
                if is_within(journey.bbox, net['bbox']):
                    candidates.append((k, net))

            if candidates != []:
                logger.debug(f'{journey.gps_hash}: Using a larger network rather than generating')
                selection = sorted(
                    candidates,
                    key=lambda x: area_from_coords(x[1]['bbox'])
                )[0]

                self.load_networks(network_id=selection[0])

                return self.networks[selection[0]]

            # see if we can use a place
            if place_cache.get_by_bbox(journey.bbox) is not None:
                bbox = place_cache.get_by_bbox(journey.bbox)['bbox']

                network = ox.graph_from_bbox(
                    bbox['north'],
                    bbox['south'],
                    bbox['east'],
                    bbox['west'],
                    network_type='all',
                    simplify=True
                )

                self.data[get_graph_id(network)] = {
                    'gps_hash': journey.gps_hash,
                    'bbox': {
                        'north': bbox['north'],
                        'south': bbox['south'],
                        'east': bbox['east'],
                        'west': bbox['west'],
                    }
                }
                self.networks[get_graph_id(network)] = network
                self.save()
                return network

        for net_id, net in self.data.items():
            if journey.gps_hash == net['gps_hash']:
                self.load_networks(network_id=net_id)
                return self.networks[net_id]

        return None

    def set(self, journey, network: MultiDiGraph, skip_save=False):
        self.last_accessed = datetime.datetime.utcnow()

        self.load()

        self.data[get_graph_id(network)] = {
            'gps_hash': journey.gps_hash,
            'bbox': journey.bbox
        }

        # need to set networks to loaded or it may be overridden on a load
        self.load_networks()
        self.networks[get_graph_id(network)] = network

        if not skip_save:
            self.save()

    def save(self):
        if any([
            not os.path.exists(self.fp),
            len(self.data) > self.last_save_len and self.last_save_len >= 0
        ]):
            logger.debug(f'Saving cache {self.fp}')

            os.makedirs(
                os.path.join(self.dir, 'networks'),
                exist_ok=True
            )

            for k, net in self.networks.items():
                network_filepath = os.path.join(self.dir, 'networks', f'{k}.pickle')
                if not os.path.exists(network_filepath):
                    with open(network_filepath, 'wb') as network_file:
                        pickle.dump(net, network_file)

            with open(self.fp, 'wb') as network_file:
                pickle.dump(self.data, network_file)

    def load(self):
        if self.loaded:
            return

        logger.debug(f'Loading cache {self.fp}')
        if not os.path.exists(self.fp):
            os.makedirs(
                os.path.dirname(self.fp),
                exist_ok=True
            )
            self.save()

        with open(self.fp, 'rb') as network_file:
            self.data = pickle.load(network_file)

        self.loaded = True
        self.last_save_len = len(self.data)
        logger.debug('Size of network cache: %sMB', get_size(self.data) / (1000 ** 2))

    def unload(self):
        self.networks = {}
        self.networks_loaded = False

    @property
    def dir(self) -> str:
        return os.path.join(NETWORK_CACHE_DIR, settings.VERSION, self.network_type)

    @property
    def fp(self) -> str:
        return os.path.join(self.dir, f'{self.fn}')

    @property
    def since_last_accessed(self):
        return (datetime.datetime.utcnow() - self.last_accessed).total_seconds()


class GroupedNetworkCaches():

    def __init__(self, *args, **kwargs):
        assert kwargs.get('cache_type', None) is not None
        self.loaded = False
        self.data = {}
        self.last_save_len = -1
        self.cache_type = kwargs['cache_type']
        self.child_class = SingleNetworkCache

        # set up references
        self.refs = {}
        os.makedirs(self.dir, exist_ok=True)
        if not os.path.exists(self.refs_path):
            with open(self.refs_path, 'w') as refs_file:
                refs_file.write(json.dumps({}))
        with open(self.refs_path, 'r') as refs_file:
            self.refs = json.loads(refs_file.read())

        if settings.CLEAN_MEMORY:
            self.memory_cleaner()

    def get_by_id(self, graph_id):
        self.load()
        for obj in self.data.values():
            graph = obj.get_by_id(graph_id)
            if graph is not None:
                return graph

    def get_at_point(self, gps_point):
        self.load()
        network_at_point = None
        for v in self.data.values():
            p = v.get_at_point(gps_point)
            if p:
                network_at_point = p
        return network_at_point

    def get(self, obj):
        self.load()
        # if journey also do gps_hash
        from via.models.journey import Journey
        from via.models.journeys import Journeys
        if isinstance(obj, (Journey, Journeys)):
            for k, v in self.data.items():
                result = v.get(obj)
                if result is not None:
                    return result
        else:
            if obj in self.refs and self.refs[obj] in self.data:
                return self.data[self.refs[obj]].get(obj)
        return None

    def set(self, k, v, skip_save=False):
        x = round(
            min(
                [d['x'] for d in dict(v.nodes(data=True)).values()]
            ),
            2
        )
        y = round(
            min(
                [d['y'] for d in dict(v.nodes(data=True)).values()]
            ),
            2
        )
        fn = '%s_%s' % (x, y)
        if fn not in self.data:
            self.data[fn] = self.child_class(network_type=self.cache_type, fn=fn)
        self.data[fn].set(k, v)
        len_refs = len(self.refs)

        # k is collisions, so needs to be serializable
        self.refs[k.gps_hash] = fn

        if len_refs < len(self.refs):
            if not skip_save:
                self.data[fn].save()
                self.save_refs()

    def save_refs(self):
        with open(self.refs_path, 'w') as refs_file:
            refs_file.write(json.dumps(self.refs))

    def load(self):
        """
        Load all the caches of this type, note that this does not load
        data, it loads only references so that things can be lazily loaded
        """
        if self.loaded:
            return

        logger.debug('loading %s', self.cache_type)

        try:
            for c in self.caches:
                self.data[c.fn] = c
        except:
            self.data = {}

        self.loaded = True

    def memory_cleaner(self):
        """
        Clean up memory from caches that haven't been used in a while
        """
        logger.debug('Cleaning memory: %s', self.cache_type)

        if logger.level <= logging.DEBUG:
            # since get_size is a little slow, don't want to waste time if
            # we're not going to log about it
            try:
                initial_memory = get_size(self)
            except:
                initial_memory = -1

        # TODO: locking
        for v in self.data.values():
            if v.since_last_accessed > 60:
                v.unload()

        if logger.level <= logging.DEBUG:
            try:
                post_memory = get_size(self)
            except:
                post_memory = -1
            logger.debug(
                'Cleaned memory, reduced %s by %s%% (%s -> %s)',
                self.cache_type,
                ((float(post_memory) - initial_memory) / initial_memory) * 100,
                initial_memory,
                post_memory
            )

        cleaner = threading.Timer(
            60 * 2,
            self.memory_cleaner
        )
        cleaner.daemon = True
        cleaner.start()

    @property
    def caches(self):
        """
        Get all the caches within the directory of this cache type
        """

        return [
            self.child_class.from_file(
                cache_type=self.cache_type,
                filepath=os.path.join(self.dir, f),
                load=False
            ) for f in os.listdir(self.dir) if f not in {'refs.json', 'networks'}
        ]

    @property
    def dir(self) -> str:
        return os.path.join(NETWORK_CACHE_DIR, settings.VERSION, self.cache_type)

    @property
    def refs_path(self) -> str:
        """
        Get the reference file path for this type of cache
        """
        return os.path.join(self.dir, 'refs.json')


class NetworkCache():

    def __init__(self):
        self.network_caches = {}

    def get_at_point(self, key, gps_point):
        if key not in self.network_caches:
            self.network_caches[key] = GroupedNetworkCaches(cache_type=key)
        return self.network_caches[key].get_at_point(gps_point)

    def get_by_id(self, graph_id):
        for cache in self.network_caches.values():
            network = cache.get_by_id(graph_id)
            if network is not None:
                return network['network']

    def get(self, key: str, journey) -> MultiDiGraph:
        if key not in self.network_caches:
            self.network_caches[key] = GroupedNetworkCaches(cache_type=key)
        return self.network_caches[key].get(journey)

    def set(
        self,
        key: str,
        journey,
        network: MultiDiGraph,
        skip_save=False
    ):
        if key not in self.network_caches:
            self.network_caches[key] = GroupedNetworkCaches(cache_type=key)
        self.network_caches[key].set(journey, network, skip_save=skip_save)

    def load(self, network_type=None):
        if network_type is not None:
            self.network_caches[network_type] = GroupedNetworkCaches(
                cache_type=network_type
            )
        else:
            networks_dir = os.path.join(NETWORK_CACHE_DIR, settings.VERSION)
            for net_type in os.listdir(networks_dir):
                self.network_caches[net_type] = GroupedNetworkCaches(
                    cache_type=net_type
                )
                self.network_caches[net_type].load()

    def save(self):
        for k, v in self.network_caches.items():
            v.save()


network_cache = NetworkCache()



import json

from unittest import TestCase, skip

from via.models.journey import Journey
from via.models.frame import Frame

from via.edge_cache import get_edge_data


class EdgeCacheTest(TestCase):

    def setUp(self):

        with open('test/resources/just_route.json') as json_file:
            self.test_data = json.load(json_file)

        self.test_journey = Journey()
        for d in self.test_data:
            self.test_journey.append(
                Frame(
                    d['time'],
                    {'lat': d['lat'], 'lng': d['lng']},
                    1  # acceleration, don't really care at the mo
                )
            )

    def test_get(self):
        self.assertEqual(
            get_edge_data(389281, 135109553, graph=self.test_journey.graph),
            {0: {'osmid': 14039949, 'oneway': True, 'name': 'York Street', 'highway': 'unclassified', 'maxspeed': '30', 'length': 34.217}}
        )

    @skip('todo')
    def test_caching(self):
        # Make sure it just does get_edge_data once
        pass



from setuptools import (
    find_packages,
    setup
)

INSTALL_REQUIRES = (
    'road_collisions',
    'cached_property',
    'backoff',
    'packaging',
    'reverse_geocoder',
    'fast_json',
    'boto3',
    'backoff',
    'haversine',
    'osmnx',
    'scikit-learn',
    'requests',
    'geopandas',
    'imagehash',
    'pillow'
)

setup(
    name='via-api',
    version='0.0.27',
    python_requires='>=3.6,<3.10',
    description='Analysing and serving crowdsourced road quality data',
    author='Robert Lucey',
    url='https://github.com/RobertLucey/via',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=INSTALL_REQUIRES,
    entry_points={
        'console_scripts': [
            'pull_journeys = via.bin.pull_journeys:main',
            'plot_journeys = via.bin.plot_journeys:main',
            'road_coverage = via.bin.road_coverage:main',
            'condition_by_street = via.bin.condition_by_street:main',
            'generate_geojson = via.bin.generate_geojson:main'
        ]
    }
)



from road_collisions.constants import COUNTY_MAP

from via import logger
from via.models.collisions.collision import Collisions


def get_collisions():
    return Collisions.load_all()


def generate_geojson():
    all_collisions = get_collisions()

    for county in COUNTY_MAP.values():
        for vehicle_type in {None, 'bicycle', 'car', 'bus'}:

            # get more specific as time goes on for fewer networks to cache
            for year in [None] + sorted(list(range(2005, 2017)), reverse=True):

                filters = {
                    'county': county.lower(),
                    'year': year,
                    'vehicle_type': vehicle_type
                }

                filters = {
                    k: v for k, v in filters.items() if v is not None and v != 'all'
                }

                logger.debug('Process collisions geojson: %s', filters)
                collisions = all_collisions.filter(
                    **filters
                )
                collisions.geojson



import datetime
from dateutil.parser import parse
from functools import lru_cache
import statistics
from collections import defaultdict
from packaging import version

from cached_property import cached_property
import geopandas as gpd
import fast_json

from shapely.ops import cascaded_union

import networkx as nx
import osmnx as ox

from via import settings
from via import logger
from via.utils import (
    window,
    get_combined_id,
    get_edge_colours
)
from via.nearest_edge import nearest_edge
from via.constants import POLY_POINT_BUFFER
from via.models.point import FramePoint, FramePoints
from via.models.frame import Frame
from via.edge_cache import get_edge_data
from via.network_cache import network_cache
from via.models.journey_mixins import (
    SnappedRouteGraphMixin,
    GeoJsonMixin,
    BoundingGraphMixin
)


class Journey(
    FramePoints,
    SnappedRouteGraphMixin,
    GeoJsonMixin,
    BoundingGraphMixin
):

    def __init__(self, *args, **kwargs):
        """

        :kwarg is_culled: If the journey is culled or not
        :kwarg transport_type: What transport type being used, defaults
            to settings.TRANSPORT_TYPE
        :kwarg suspension: If using suspension or not, defaults
            to settings.SUSPENSION
        """
        self.gps_inclusion_iter = 0
        self.too_slow = False

        data = []
        if 'data' in kwargs:
            data = kwargs.pop('data')

        kwargs.setdefault('child_class', FramePoint)
        super().__init__(*args, **kwargs)

        self.extend(data)

        self._version = kwargs.get('version', None)

        self.is_culled = kwargs.get('is_culled', False)
        self.is_sent = kwargs.get('is_sent', False)

        self.transport_type = str(kwargs.get('transport_type', 'unknown')).lower()
        self.suspension = kwargs.get('suspension', None)

        self.network_type = kwargs.get('network_type', 'all')
        self._timestamp = kwargs.get('timestamp', None)

        self.included_journeys = []

        self.last_gps = None

    def __del__(self):
        attrs_to_del = [
            'edge_quality_map',
            'route_graph'
        ]

        for attr in attrs_to_del:
            try:
                delattr(self, attr)
            except:
                pass

    @staticmethod
    def parse(objs):
        if isinstance(objs, Journey):
            return objs

        if isinstance(objs, dict):
            return Journey(
                **objs
            )

        raise NotImplementedError(
            f'Can\'t parse journey from type {type(objs)}'
        )

    @staticmethod
    @lru_cache(maxsize=500)
    def from_file(filepath: str):
        """
        Given a file get a Journey object

        :param filepath: Path to a saved journey file
        :rtype: via.models.journey.Journey
        """
        logger.debug(f'Loading journey from {filepath}')

        # TODO: should cache stages of processed files from raw (currently,
        # which uses frames instead of frame points) to processed (which
        # it will use frame points) so we can skip the building up of context
        # when we should be able to find from a raw file path

        with open(filepath, 'r') as journey_file:
            return Journey(
                **fast_json.loads(journey_file.read())
            )

    def set_contexts(self):
        for (one, two, three, four, five, six, seven) in window(self, window_size=7):
            if not four.is_context_populated:
                four.set_context(
                    pre=[one, two, three],
                    post=[five, six, seven]
                )

    def extend(self, objs):
        for obj in objs:
            self.append(obj, set_contexts=False)
        self.set_contexts()

    def append(self, obj, set_contexts=True):
        """
        NB: appending needs to be chronological (can be reversed, just so
        long as it's consistent) as if no accelerometer data it assigns
        the accelerometer data to the previously seen gps

        Though journey data may not have time it will (should) always be
        chronological
        """
        # TODO: warn if not chronological
        if isinstance(obj, FramePoint):
            self._data.append(
                obj
            )
        else:
            # Most datapoints are only accelerometer so we need to find the
            # closest point with gps in the past to add the accelerometer
            # data to

            frame = Frame.parse(obj)

            frame_gps_populated = frame.gps.is_populated

            if frame_gps_populated:
                if self.gps_inclusion_iter % settings.GPS_INCLUDE_RATIO == 0:
                    self.last_gps = frame.gps
                else:
                    frame.gps = self.last_gps
                self.gps_inclusion_iter += 1
            else:
                if not hasattr(self, 'last_gps'):
                    return
                if self.last_gps is None:
                    return
                frame.gps = self.last_gps

            if len(self._data) == 0:
                self._data.append(
                    FramePoint(frame.time, frame.gps, frame.acceleration)
                )
                return

            # Remove points that are too slow / fast in relation to
            # the previous point
            if frame_gps_populated or frame.gps.is_populated:
                metres_per_second = self._data[-1].speed_between(frame)
                if metres_per_second is not None:
                    if any([
                        metres_per_second < settings.MIN_METRES_PER_SECOND,
                        metres_per_second > settings.MAX_METRES_PER_SECOND
                    ]):
                        self.too_slow = True
                    else:
                        self.too_slow = False

            if self.too_slow:
                # If we want real speed we cannot return here without
                # appending the frame
                # possibly add all and annotate to skip due to speed
                return

            if self._data[-1].gps == frame.gps:
                self._data[-1].append_acceleration(frame.acceleration)
            else:
                self._data.append(
                    FramePoint(frame.time, frame.gps, frame.acceleration)
                )

        if set_contexts:
            self.set_contexts()

    def get_indirect_distance(self, n_seconds=10):
        """
        NB: Data must be chronological

        :param n_seconds: use the location every n seconds as if the
            location is calculated too frequently the distance
            travelled could be artificially inflated
        :rtype: float
        :return: distance travelled in metres
        """
        previous_used_frame = None
        distances = []

        for frame in self:
            if previous_used_frame is None:
                previous_used_frame = frame
            elif frame.time >= previous_used_frame.time + n_seconds:
                distances.append(
                    previous_used_frame.distance_from(
                        frame
                    )
                )
                previous_used_frame = frame

        return sum(distances)

    def get_avg_speed(self, n_seconds=30):
        """
        NB: Data must be chronological

        :param n_seconds: use the location every n seconds as if the
                        location is calculated too frequently the distance
                        travelled could be artificially inflated
        :rtype: float
        :return: avg speed in metres per second
        """
        return self.get_indirect_distance(n_seconds=n_seconds) / self.duration

    def serialize(self, minimal=False, include_time=True, include_context=True):
        data = {
            'uuid': str(self.uuid),
            'version': str(self.version),
            'data': super().serialize(include_time=include_time, include_context=include_context),
            'transport_type': self.transport_type,
            'suspension': self.suspension,
            'is_culled': self.is_culled,
            'is_sent': self.is_sent
        }

        if minimal is False:
            data.update(
                {
                    'direct_distance': self.direct_distance,
                    'indirect_distance': {
                        1: self.get_indirect_distance(n_seconds=1),
                        5: self.get_indirect_distance(n_seconds=5),
                        10: self.get_indirect_distance(n_seconds=10),
                        30: self.get_indirect_distance(n_seconds=30)
                    },
                    'data_quality': self.data_quality,
                    'duration': self.duration,
                    'avg_speed': self.get_avg_speed()
                }
            )

        return data

    def plot_route(
        self,
        use_closest_edge_from_base=False,
        colour_map_name='bwr',
        plot_kwargs={}
    ):
        """

        :kwarg use_closest_from_base: For each point on the actual route,
            for each node use the closest node from the original base graph
            the route is being drawn on
        :kwarg colour_map_name:
        :kwarg plot_kwargs: A dict of kwargs to pass to whatever plot is
            being done
        """
        base = self.graph
        if use_closest_edge_from_base:
            edge_colours = get_edge_colours(
                base,
                colour_map_name,
                edge_map=self.edge_quality_map
            )

        else:
            base.add_nodes_from(self.route_graph.nodes(data=True))
            base.add_edges_from(self.route_graph.edges(data=True))

            edge_colours = get_edge_colours(
                base,
                colour_map_name,
                key_name='avg_road_quality'
            )

        ox.plot_graph(
            base,
            edge_color=edge_colours,
            **plot_kwargs
        )

    @property
    def timestamp(self):
        if self._timestamp is None:
            # FIXME: We shouldn't need to do this but the ui always includes earliest / latest as a filter
            return datetime.datetime(2021, 1, 1)
        return parse(self._timestamp)

    @cached_property
    def edge_quality_map(self):
        """
        Get a map between edge_hash and road quality of the road. edge_map
        being edge_id and road quality being something that hasn't been
        defined yet TODO

        :rtype: dict
        """

        data = {}
        for edge_id, single_edge_data in self.edge_data.items():
            qualities = [edge['avg_road_quality'] for edge in single_edge_data]
            if len(qualities) == 0:
                data[edge_id] = {
                    'avg': 0,
                    'count': 0,
                    'speed': 0
                }
            else:
                data[edge_id] = {
                    'avg': int(
                        statistics.mean(
                            qualities
                        )
                    ),
                    'count': len(qualities),
                    'speed': 0
                }

        return {
            edge_id: d for edge_id, d in data.items() if d['count'] >= settings.MIN_PER_JOURNEY_USAGE
        }

    @property
    def edge_data(self):
        """

        :rtype: dict
        :return: {edge_id: [{edge_data}, {edge_data}]}
        """
        data = defaultdict(list)
        bounding_graph = self.graph
        route_graph = self.route_graph

        nearest_edge.get(bounding_graph, self._data)

        raw_edges_list = []
        for (our_origin, our_destination) in window(self, window_size=2):
            nearest_edges = nearest_edge.get(
                bounding_graph,
                [
                    our_origin
                ]
            )[0]

            edge = our_origin.get_best_edge(
                nearest_edges,
                mode=settings.NEAREST_EDGE_METHOD,
                graph=bounding_graph
            )

            our_edge_data = get_edge_data(
                our_origin.uuid,
                our_destination.uuid,
                graph=route_graph
            )

            raw_edges_list.append(
                [
                    get_combined_id(edge[0][0], edge[0][1]),
                    our_edge_data
                ]
            )

        edges_list = []
        for (pre, current, post) in window(raw_edges_list, window_size=3):
            if pre[0] == post[0] and current[0] != pre[0]:
                current[0] = pre[0]
            edges_list.append(current)

        for edge in edges_list:
            data[edge[0]].append(
                edge[1]
            )

        return data

    @cached_property
    def route_graph(self):
        """
        Get a graph of the journey without snapping to closest node / edge
        """
        graph = nx.Graph()

        combined_edge_data = defaultdict(list)

        for (origin, destination) in window(self, window_size=2):
            edge_id = get_combined_id(origin.uuid, destination.uuid)

            graph.add_node(
                origin.uuid,
                **{'x': origin.gps.lng, 'y': origin.gps.lat}
            )
            graph.add_node(
                destination.uuid,
                **{'x': destination.gps.lng, 'y': destination.gps.lat}
            )

            distance = origin.distance_from(destination)
            combined_edge_data[edge_id].append(
                {
                    'origin': origin,
                    'destination': destination,
                    'distance': distance,
                    'road_quality': origin.road_quality,
                    'speed': 0
                    # TODO: other bits, speed / elevation maybe?
                }
            )

        merged_edge_data = {}
        for shared_id, values in combined_edge_data.items():

            merged_edge_data[shared_id] = {
                'origin': values[0]['origin'],
                'destination': values[0]['destination'],
                'distance': values[0]['distance'],
                'avg_road_quality': statistics.mean([val['road_quality'] for val in values]),
                'max_road_quality': max([val['road_quality'] for val in values]),
                'speed': 0
            }

        for shared_id, values in merged_edge_data.items():
            graph.add_edge(
                values['origin'].uuid,
                values['destination'].uuid,
                length=values['distance'],
                avg_road_quality=values['avg_road_quality'],
                max_road_quality=values['max_road_quality']
            )

        return graph

    @property
    def bbox(self):
        return {
            'north': self.most_northern,
            'south': self.most_southern,
            'east': self.most_eastern,
            'west': self.most_western
        }

    @property
    def poly_graph(self):
        """
        Get a graph of the journey but excluding nodes far away from the route

        :rtype: networkx.classes.multidigraph.MultiDiGraph
        """

        if network_cache.get('poly', self) is None:
            logger.debug(f'poly > {self.gps_hash} not found in cache, generating...')

            # TODO: might want to not use polygon for this since we could
            # get the benefits of using a parent bbox from the cache

            # Maybe use city if possible and then truncate_graph_polygon
            points = self.get_multi_points()

            buf = points.buffer(POLY_POINT_BUFFER, cap_style=3)
            boundary = gpd.GeoSeries(cascaded_union(buf))

            network = ox.graph_from_polygon(
                boundary.geometry[0],
                network_type=self.network_type,
                simplify=True
            )

            # TODO: might want to merge our edge_quality_data with
            # edge data here

            network_cache.set('poly', self, network)

        return network_cache.get('poly', self, poly=True)

    @property
    def graph(self):
        """
        Get a graph of the journey but excluding nodes far away from the route

        :rtype: networkx.classes.multidigraph.MultiDiGraph
        """
        return self.bounding_graph

    @property
    def all_points(self):
        """
        Return all the points in this journey.

        May be a bit expensive since it sets contexts. Can probably get away
        with selectively doing this
        """
        self.set_contexts()
        return self._data

    @property
    def version(self):
        """
        Get the version of the app used to generate this data

        :rtype: version.Version
        """
        if isinstance(self._version, version.Version):
            return self._version
        if isinstance(self._version, type(None)):
            return version.parse('0.0.0')
        return version.parse(self._version)



import pickle

from unittest import TestCase
from mock import patch

from via.place_cache import PlaceCache


def get_cork(*args, **kwargs):
    with open('test/resources/cork_graph.pickle', 'rb') as f:
        return pickle.load(f)


class PlaceCacheTest(TestCase):

    @patch('osmnx.graph_from_place', get_cork)
    def test_get(self):
        # might want to mock ox.graph_from_place to get from wherever in test resources
        cache = PlaceCache()
        self.assertFalse('cork ireland' in cache.data)
        cork = cache.get('Cork, Ireland')
        self.assertEqual(
            cork,
            {
                'east': -8.3567461,
                'north': 51.9623212,
                'south': 51.8275379,
                'west': -8.6368609
            }
        )
        self.assertTrue('cork ireland' in cache.data)

    def test_get_by_bbox_within(self):
        cache = PlaceCache()
        self.assertIsNotNone(
            cache.get_by_bbox(
                {
                    'north': 53.61,
                    'south': 53.4,
                    'east': -6.2,
                    'west': -6.3,
                }
            )
        )

    def test_get_by_bbox_exact(self):
        cache = PlaceCache()
        self.assertIsNotNone(
            cache.get_by_bbox(
                {
                    'north': 53.626487,
                    'south': 53.3018049,
                    'east': -6.1366563,
                    'west': -6.433500
                }
            )
        )

    def test_get_by_bbox_nowhere(self):
        cache = PlaceCache()
        self.assertIsNone(
            cache.get_by_bbox(
                {
                    'north': 0,
                    'south': 0,
                    'east': 0,
                    'west': 0,
                }
            )
        )



import os
import operator

import fast_json

from via import logger
from via.constants import GEOJSON_DIR
from via.utils import (
    get_journeys,
    should_include_journey
)
from via.models.journeys import Journeys
from via.geojson.utils import generate_basename


def get_generation_config(transport_type=None, version=None, version_op=None, earliest_time=None, latest_time=None, place=None):
    config = []
    if transport_type in {None, 'all'}:
        config = [
            {'transport_type': 'all', 'name': 'all', 'version': version, 'version_op': version_op if version else None, 'earliest_time': earliest_time, 'latest_time': latest_time, 'place': place},
            {'transport_type': 'bike', 'name': 'bike', 'version': version, 'version_op': version_op if version else None, 'earliest_time': earliest_time, 'latest_time': latest_time, 'place': place},
            {'transport_type': 'car', 'name': 'car', 'version': version, 'version_op': version_op if version else None, 'earliest_time': earliest_time, 'latest_time': latest_time, 'place': place},
            {'transport_type': 'bus', 'name': 'bus', 'version': version, 'version_op': version_op if version else None, 'earliest_time': earliest_time, 'latest_time': latest_time, 'place': place}
        ]
    else:
        config = [
            {'transport_type': transport_type, 'name': transport_type, 'version': version, 'version_op': version_op if version else None, 'earliest_time': earliest_time, 'latest_time': latest_time, 'place': place}
        ]

    return config


def generate_geojson(
    transport_type,
    version=False,
    version_op=None,
    earliest_time=None,
    latest_time=None,
    place=None
):

    os.makedirs(
        os.path.join(GEOJSON_DIR),
        exist_ok=True
    )

    for config_item in get_generation_config(
        transport_type=transport_type,
        version=version,
        version_op=version_op,
        earliest_time=earliest_time,
        latest_time=latest_time,
        place=place
    ):
        logger.info(f'Generating geojson for "{config_item["transport_type"]}"')

        journeys = get_journeys(
            transport_type=config_item['transport_type'],
            earliest_time=earliest_time,
            latest_time=latest_time,
        )

        basename = generate_basename(
            name=config_item['name'],
            version=config_item['version'],
            version_op=config_item['version_op'],
            earliest_time=config_item['earliest_time'],
            latest_time=config_item['latest_time'],
            place=config_item['place']
        )
        geojson_file = os.path.join(
            GEOJSON_DIR,
            f'{basename}.geojson'
        )

        journeys = Journeys(
            data=[
                journey for journey in journeys if should_include_journey(
                    journey,
                    version_op=getattr(operator, version_op) if version_op is not None else None,
                    version=version
                )
            ]
        )

        with open(geojson_file, 'w') as json_file:
            fast_json.dump(
                journeys.geojson,
                json_file
            )
